from dsframework.base.trainer.test import ZIDSTrainerTest
from pl_wrapper.plmodel import generatedProjectNamePlmodel
from dsframework.base.trainer.pl_wrapper import *

##
# @file
# @brief generatedClass class, executes test process.
class generatedClass(ZIDSTrainerTest):
    """! Test class, runs the model against a labeled test dataset to evaluate its metrics.

        Three things needs to be done for this to work:
            1. Receive data_module with labeled dataset loaded.
            2. Load model
            3. Create and Run trainer instance.
    """

    def __init__(self, data_module=None, model_path=None, trainer_config=None):
        """generatedClass Initializer.

        Runs trainer_setup() and load_model() methods/
        """
        super().__init__(data_module, model_path, trainer_config)
        self.trainer_setup(trainer_config)
        self.load_model(model_path)

    def trainer_setup(self, trainer_config):
        """! Instantiate trainer class

        Needs to be implemented.

        Example (pytorch lightning):
            self.trainer = pl.Trainer(**trainer_config)
        """

        if self.trainer is None:
            raise Exception('trainer_setup() method in test.py, not implement.')

    def load_model(self, model_path):
        """! Loads a model

        Needs to be implemented.

        Example (pytorch lightning):
            self.model = generatedProjectNamePlmodel.load_from_checkpoint(model_path)
        """

        if self.model is None:
            raise Exception('load_model() method in test.py, not implement.')

    def execute_test(self):
        """! Implement test process and return the results.

        Returns:
            test results - Metric results from test.

        Example (pytorch lightning):
            if self.model is None or self.data_module is None or self.trainer is None:
                return None
            self.trainer.test(self.model, self.data_module)
            return self.model.get_test_results()
        """

        raise Exception('Methods execute_test() not implemented in test.py.')



import os

from trainer.data_module import generatedProjectNameDataModule
from trainer.network_module import generatedProjectNameNetworkModule
from trainer.model import generatedProjectNameModel
from dsframework.base.trainer.pl_wrapper import *


def train(config, train_config):

    mnist_full = datasets.MNIST(os.getcwd() + '/datasets', download=True, train=True, transform=transforms.ToTensor())
    mnist_test = datasets.MNIST(os.getcwd() + '/datasets', download=True, train=False, transform=transforms.ToTensor())

    mnist_data_module = generatedProjectNameDataModule(
        batch_size_train=config['batch_size_train'],
        batch_size_eval=config['batch_size_eval'],
        dataset=mnist_full,
        test_set=mnist_test,
        num_workers=config['num_workers'],
        seed=config['seed'],
        split_percent=0.8
    )

    model = generatedProjectNameModel(
        loss_function=config['loss_function'],
        num_classes=config['num_classes'],
        learn_rate=config['learning_rate'],
        nnetwork=config['nnetwork']
    )

    # automatically restores model, epoch, step, LR schedulers, apex, etc...
    # trainer.fit(model, ckpt_path="some/path/to/my_checkpoint.ckpt")
    trainer = pl.Trainer(**train_config)
    trainer.fit(model, mnist_data_module)

    trainer.test(model, mnist_data_module)


# saves top-K checkpoints based on "val_loss" metric
checkpoint_callback = ModelCheckpoint(
    save_top_k=10,
    save_last=True,
    monitor="val_loss",
    mode="max",
    filename="my-chkpnt-{epoch:02d}-{val_loss:.2f}",
)

if __name__ == '__main__':

    general_config = {
        'learning_rate': 1e-3,
        'loss_function': F.cross_entropy,
        'batch_size_train': 32,
        'batch_size_eval': 32,
        'num_classes': None,
        'num_workers': 1,
        'seed': 42,
        'nnetwork': generatedProjectNameNetworkModule.nnetwork
    }

    trainer_config = {
        'max_epochs': 10,
        'accelerator': None,  # 'cpu', 'gpu', 'tpu', 'ipu', 'auto', None
        'devices': None,  # number of accelerators: -1-all, 1, 3, 4...8, None, [1, 3]-gpu ids, None
        'strategy': None,  # 'dp', 'ddp', 'ddp2', 'ddp_spawn'
        'num_processes': None,  # number of cpus
        'enable_checkpointing': True,
        'default_root_dir': './',  # root dir to save checkpoints
        'callbacks': [checkpoint_callback],
        'enable_progress_bar': False
    }

    train(general_config, trainer_config)



import subprocess
import os
import sys
from json import dumps as json_dumps
from json import load as json_load
from json import loads as json_loads
from yaml import dump as yaml_dump

from datetime import timedelta
from datetime import datetime

from google.cloud.dataproc_v1 import ClusterControllerClient
from google.cloud.dataproc_v1 import WorkflowTemplatePlacement
from google.cloud.dataproc_v1.types import OrderedJob
from google.cloud.dataproc_v1.types import PySparkJob
from google.cloud.dataproc_v1.services.workflow_template_service.async_client import WorkflowTemplateServiceClient
from google.cloud.dataproc_v1.types.workflow_templates import WorkflowTemplate


##
# @file
# @brief Defines workflow base class.
class ZIDS_Workflow():
    def __init__(self, pip_packages):
        """! ZIDS_Workflow initializer.
        Loads config file and initializes parameters.

            Args:
                 pip_packages: A list of pip packages read from requirements file.
        """

        file_path = sys.modules[self.__class__.__module__].__file__
        curr_path = file_path[:file_path.rfind("batch_files")]

        with open(curr_path + '/batch_files/batch_config.json') as batch_cfg_file:
            self.config = json_load(batch_cfg_file)

        self.user_email = ''
        env_type = os.environ.get('SPRING_PROFILES_ACTIVE')
        if (not env_type == 'production') and (not env_type == 'staging'):
            command = 'gcloud config get-value account'
            user_email = subprocess.check_output(command, shell=True).decode(sys.stdout.encoding)
            self.user_email = '/' + user_email.split("@")[0]

        if self.user_email:
            self.config['user_email'] = self.user_email

        self.default_args = [json_dumps(self.config),
                             "start_date_placeholder", "end_date_placeholder", "params_placeholder"]

        self.region = self.config['region']
        self.zone = self.region + '-b'
        self.project_id = self.config['project_id']
        self.template_id = self.config['template_id']
        self.bucket_name = self.config['bucket_name']
        self.project_name = self.config['project_name']
        self.autoscaling_policy_id = self.project_name + '-auto-scaling-policy'

        self.unique_iteration_id = self.config['unique_iteration_id']
        self.unique_template_id = self.config['unique_template_id']
        if self.unique_template_id:
            self.template_id = self.template_id + '_' + self.unique_template_id

        if self.unique_iteration_id:
            self.folder_path = self.project_name + self.user_email + '/unique_iteration_id_' + self.unique_iteration_id
        else:
            self.folder_path = self.project_name + self.user_email + '/main'

        self.bucket_path = f'gs://{self.bucket_name}/{self.folder_path}'
        self.project_path = 'projects/{project_id}/regions/{region}'.format(project_id=self.project_id,
                                                                            region=self.region)

        if self.config['cluster_conf']['managed_cluster']:
            # Configure pip packages
            self.config['cluster_conf']['managed_cluster']['config']['gce_cluster_config']['metadata']['PIP_PACKAGES'] \
                = pip_packages

            # Configure initialization script
            cluster_init_action = {"executable_file": f"{self.bucket_path}"
                                                      f"/batch_files/scripts/cluster-init-actions.sh"}
            cluster_init_list = self.config['cluster_conf']['managed_cluster']['config']['initialization_actions']

            if not isinstance(cluster_init_list, list):
                cluster_init_list = []

            cluster_init_list.append(cluster_init_action)

        # Define workflow client
        self.serv_client = WorkflowTemplateServiceClient(
            client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
        )

        # Define extra stage parameters
        self.extra_params = json_dumps(self.config['extra_job_params'])

    def check_cluster(self):
        """! ZIDS_Workflow get_cluster.
        Tries to locate existing cluster that should be set up by this point.
        """
        clusterClient = ClusterControllerClient(
            client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
        )
        cluster_name = self.config['cluster_conf']['managed_cluster']['cluster_name']
        try:
            exist_cluster = clusterClient.get_cluster(
                project_id=self.project_id,
                region=self.region,
                cluster_name=cluster_name,
            )
            if exist_cluster:
                del self.config['cluster_conf']['managed_cluster']
                self.config['cluster_conf']['cluster_selector'] = {
                    "cluster_labels": {
                        "goog-dataproc-cluster-name": cluster_name
                    }
                }
        except Exception as e:
            if "404" in str(e):
                print(f"Cluster {cluster_name} not found. Will be created when workflow is instantiated")
            else:
                # If there's something other than 404, we want to know, but the flow can carry on.
                print(f'Error getting cluster: {e}. Cluster will be created on demand')

    def create_template_structure(self) -> WorkflowTemplate:
        """! ZIDS_Workflow create_template_structure.
        Tries to create a workflow template, containing all available stages.

            Returns:
                WorkflowTemplate() - Created template with basic arguments
        """
        cluster_config = WorkflowTemplatePlacement(self.config['cluster_conf'])

        # if results not exist
        template = WorkflowTemplate()
        template.id = self.template_id
        template.name = self.project_path + '/workflowTemplates/' + template.id
        template.placement = cluster_config
        return template

    def create_ordered_job(self, stage_dir: str, spark_properties, python_files, archive_files,
                           additional_args=[], additional_jars=[]) -> OrderedJob:
        """! ZIDS_Workflow create_ordered_job.
        Create an ordered job from a stage directory

            Args:
                stage_dir: Stage directory, should hold the main.py file and configuration
                           example: batch_files/stages/stage_write_page_list_to_big_query
                spark_properties: A list, containing different spark properties
                python_files: A list of files which hold the project code (Can be archived)
                archive_files: A list of compressed files that the stage requires
                additional_args: A list of additional arguments, to add on top of the default ones
                additional_jars: A list of additional jar files, to add on top of the default ones
            Returns:
                An ordered job
        """

        # Assume that the last part of the path is the stage name
        stage_name = stage_dir.rsplit('/', 1)[-1]

        created_stage = OrderedJob()
        pyJob = PySparkJob()
        pyJob.main_python_file_uri = f'{self.bucket_path}/{stage_dir}/main.py'
        pyJob.archive_uris = [f'{self.bucket_path}/dist/{file_name}' if file_name.find('gs:') == -1
                              else file_name for file_name in archive_files]
        pyJob.python_file_uris = [f'{self.bucket_path}/dist/{file_name}' if file_name.find('gs:') == -1
                                  else file_name for file_name in python_files]
        pyJob.properties = spark_properties
        pyJob.args = self.default_args + additional_args
        pyJob.jar_file_uris = self.default_jars + additional_jars
        created_stage.step_id = f'{self.project_name}-stage_{stage_name}'
        created_stage.pyspark_job = pyJob
        return created_stage

    def check_template(self):
        """! Check for an existing template. If found, delete it.
        """

        exist_template = None
        exist_template = self.get_workflow_template()

        # Deleting template if it exists
        if exist_template:
            self.delete_workflow_template(exist_template)

    def create_template(self, template):
        """! Create the complete template on the remote location.
            Args:
                template: WorkflowTemplate()
        """
        try:
            result = self.serv_client.create_workflow_template(
                parent=self.project_path,
                template=template
            )
            print(f'template {template.id} created successfully')
        except Exception as e:
            print(f'Error creating template: {e}')

    def get_workflow_template(self) -> WorkflowTemplate:
        """! Get the current template
            Returns:
                WorkflowTemplate, if such was retrieved

        """

        template_name = self.project_path + '/workflowTemplates/' + self.template_id

        try:
            existing_template = self.serv_client.get_workflow_template(
                name=template_name,
            )
            print(f'template {self.template_id} retrieved successfully')
        except Exception as e:
            print(f'Cannot get template, exception: {e}')
            return None

        return existing_template

    def instantiate_template(self, template):
        """! Instantiate the template into a working workflow, overriding DAG
        Do this only on dev mode
            Args:
                template: WorkflowTemplate()
        """

        start_date = datetime.now().replace(second=0, microsecond=0)
        end_date = start_date + timedelta(days=1)
        start_date_str = start_date.strftime("%Y-%m-%dT%H:%M:%S+00:00")
        end_date_str = end_date.strftime("%Y-%m-%dT%H:%M:%S+00:00")

        try:
            self.serv_client.instantiate_workflow_template(
                name=template.name,
                parameters={"START_DATE": start_date_str, "END_DATE": end_date_str, "PARAMS": self.extra_params}
            )
            print(f'template {template.id} instantiated successfully')
        except Exception as e:
            print(f'Error instantiating template: {e}')

    def delete_workflow_template(self, template):
        """! Delete a given template
            Args:
                template: WorkflowTemplate()
        """

        try:
            self.serv_client.delete_workflow_template(
                name=template.name,
            )
            print(f'template {template.id} deleted successfully')
        except Exception as e:
            print(f'Error deleting template: {e}')

    def create_workflow_template(self):
        """! Create the workflow template, along with all relevant stages
        """
        raise NotImplementedError

    def create_yaml_file(self, workflow_template):
        meta_msg = workflow_template._meta.parent
        workflow_str = meta_msg.to_json(workflow_template)
        clean_workflow_str = workflow_str.strip()
        workflow_json = json_loads(clean_workflow_str)
        # Remove unnecessary fields which are enums that are messing up yaml parsing. Needs to be resolved later if
        # they are needed. Import the Enum and give correct values
        del workflow_json['placement']['managedCluster']['config']['gceClusterConfig']['privateIpv6GoogleAccess']
        del workflow_json['placement']['managedCluster']['config']['masterConfig']['preemptibility']
        del workflow_json['placement']['managedCluster']['config']['workerConfig']['preemptibility']

        file_path = sys.modules[self.__class__.__module__].__file__
        curr_path = file_path[:file_path.rfind("workflow")]
        with open(curr_path + '/workflow.yaml', 'w') as yml:
            yaml_dump(workflow_json, yml)



import ast

import click
import os
import shutil
import re
import sys
import platform
import subprocess
import site
import webbrowser
from datetime import datetime
from json import load as json_load
from json import dumps as json_dumps
from dsframework.documentation.scripts import doxygen_handler

isWindows = platform.system() == 'Windows'

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))

directories = {
    'main': '',
    'pipeline': '',
    'artifacts': '',
    'models': '',
    'vocabs': '',
    'other': '',
    'preprocessor': '',
    'predictables': '',
    'predictors': '',
    'forcers': '',
    'postprocessor': '',
    'schema': '',
    'tester': '',
    'test_schema': '',
    'trainer': '',
    'server': '',
    'dsp': '',
    'documentation': ''
}

##
# @file
# @brief This file defines and implements the framework CLI. It helps us to:\n
#        Create: Project/Forcer/Predictable/Tester-files/Deploy-files/Cloud-Eval-Files  \n
#        Run   : Server / Evaluation (CSV,Cloud) \n


class AliasedGroup(click.Group):
    """! AliasedGroup implements click.Group base class."""

    ##
    # @hidecallgraph @hidecallergraph
    def get_command(self, ctx, cmd_name):
        try:
            cmd_name = ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)


@click.group(cls=AliasedGroup)
@click.version_option(package_name='dsFramework')
def cli():
    """
    DS framework cli

    ** How to use **

    g = generate

    Create project
    ==============
    dsf-cli g project my-new-project

    Generate forcer
    ===============
    dsf-cli g forcer my-new-forcer

    Generate predictable
    ====================
    dsf-cli g predictable my-new-predictable

    Create files
    ===================
    dsf-cli create-files [tester / deploy / cloud_eval / documentation / batch / trainer]

    run server
    ==========
    dsf-cli run-server

    run evaluation on a csv input
    =============================
    dsf-cli evaluate input-csv-file batch-size

    ==============
    Documentation
    ==============

    Generate documentation
    ======================
    dsf-cli generate documentation full

    Generate documentation - without base classes:
    ==============================================
    dsf-cli generate documentation simple

    Show documentation
    ==================
    dsf-cli show-documentation

    Upload documentation
    ====================
    dsf-cli upload-documentation

    ========================
    Batch project abilities
    ========================

    DataProc Cluster Manipulation
    =============================
    dsf-cli dataproc-cluster <create / stop / start / delete / connect>

    Workflow Template Manipulation
    ==============================
    dsf-cli workflow-template <create / create_yaml / delete / instantiate>

    DAG Manipulation
    ================
    dsf-cli dag <create / import / delete>

    Upload files to bucket
    ======================
    dsf-cli upload-batch-files <jars, stages, whl , cluster_init, all>

    Miscellaneous
    Create sonar scanner properties file:
    ====================================
    dsf-cli create-sonar-file current_project_repo_name

    Install requirements for trainer
    ================================
    dsf-cli install-trainer-packages

    Install documentation doxygen
    =============================
    dsf-cli install-documentation-requirements

    Deploy project to GAE
    =====================
    dsf-cli deploy_service_to_gae
    """
# @click.option("--type", prompt="type", help="type of component")
# @click.option("--project_name", prompt="project_name", help="project_name")
# def apis(type, project_name):


@cli.command()
@click.argument('gen_type')
@click.argument('project_name')
def generate(gen_type, project_name):
    """! Build a generation function from the incoming gen_type.
        Args:
            gen_type: A string containing one of the following: [project / forcer / predictable / documentation]
            project_name: A string declaring our project name
    """
    try:
        f = globals()["generate_%s" % gen_type]
    except Exception as e:
        click.echo('type ' + gen_type + ' not found')
        return
    f(project_name)

@cli.command()
@click.argument('file_type')
def create_files(file_type):
    """! Build a file creation function from the incoming file type.
        Args:
            file_type: A string containing one of the following:
                [tester / deploy / cloud_eval / documentation / batch / trainer]
    """
    try:
        f = globals()["create_%s_files" % file_type]
    except Exception as e:
        click.echo('type ' + file_type + ' not found')
        return
    f()


@cli.command()
def run_server():
    """! Start a uvicorn server by running /server/main.py """
    currentPipelineFolder = os.path.abspath(os.getcwd())
    currentParentFolder = os.path.join(currentPipelineFolder,"..")
    os.environ["PYTHONPATH"] = currentPipelineFolder
    folder = currentPipelineFolder + '/server/main.py'
    subprocess.call('python ' + folder, shell=True)

@cli.command()
@click.argument('tag')
@click.argument('env', default="prd")
def cloud_eval(tag, env):
    # """Cloud evaluation running on AWS """
    currentPipelineFolder = os.path.abspath(os.getcwd())
    currentParentFolder = os.path.join(currentPipelineFolder,"..")
    os.environ["PYTHONPATH"] = currentPipelineFolder
    folder = currentPipelineFolder + '/tester/dsp/get_model_options.py'
    output = subprocess.check_output('python ' + folder + ' ' + env, shell=True, encoding='utf-8').strip()
    if output:
        # print('output', output)
        l = output.split('\n')
        m = l[-1]
        try:
            l = ast.literal_eval(m)
            o = l[0]
            model = {
                'name': o['name'],
                'id': o['id'],
                'repo': o['repo'],
            }
            folder = currentPipelineFolder + '/tester/dsp/get_dataset_options.py'
            output = subprocess.check_output('python ' + folder + ' ' + env + ' ' + model['name'] + ' ' + str(model['id']),
                                             shell=True, encoding='utf-8').strip()
            l = output.split('\n')
            m = l[-1]
            l = ast.literal_eval(m)
            selected = let_user_pick(l)
            # print('selected', selected)
            if not selected == None:
                selected_dataset = l[selected]
                # print('model',model)
                # print('selected_dataset', selected_dataset)
                folder = currentPipelineFolder + '/tester/dsp/add_experiment.py'
                subprocess.call(('dvc_push.sh'), shell=True)
                output = subprocess.call(('python', folder, env, json_dumps(model), json_dumps(selected_dataset), tag), shell=True)
                print('you can see your experiment in que')
            else:
                print('selection does not exist')
        except Exception as e:
            print(m)
    else:
        print('could not get models')


def create_deploy_files():
    """! Create deploy files (created through 'Create project', but can be used as a standalone function) """
    copy_deploy_files('')


def create_cloud_eval_files():
    """! Create cloud evaluation files (created through 'Create project', but can be used as a standalone function) """
    currentPipelineFolder = os.path.abspath(os.getcwd())
    copy_cloud_eval_files(currentPipelineFolder)


def create_tester_files():
    """! Create tester files (created through 'Create project', but can be used as a standalone function) """
    create_tester_env()


def create_documentation_files():
    """! Create documentation files (created through 'Create project', but can be used as a standalone function) """
    project_name = os.path.basename(os.getcwd())
    copy_documentation_files(os.path.join(directories['main'], 'documentation'))
    update_doc_configuration(project_name, os.path.join(directories['main'], 'documentation'))


def create_batch_files():
    """! Create batch files (created through 'Create project', but can be used as a standalone function) """
    copy_batch_files(directories['main'])


def create_trainer_files():
    """! Create trainer files (created through 'Create project', but can be used as a standalone function) """
    copy_trainer_files()


@cli.command()
@click.argument('action')
def dataproc_cluster(action: str):
    """! Handle cluster action, either create,  start, stop, delete, or connect to the Dataproc cluster
            Args:
                 action: an action to perform, chosen from: {create, start, stop, delete, connect}
    """
    valid_actions = {'create', 'start', 'connect', 'stop', 'delete'}
    if action not in valid_actions:
        print(f"Can't determine file type {action}. Please use one of: {valid_actions}")
        return

    batch_helper = config_batch_helper()

    if action == 'create':
        # Create DataProc cluster according to project specific configuration
        batch_helper.create_dataproc_cluster()
    elif action == 'start':
        # Start an existing DataProc cluster according to project specific configuration
        batch_helper.start_dataproc_cluster()
    elif action == 'connect':
        # Connect via ssh to an existing DataProc cluster according to project specific configuration.
        # Connection will be made to the master node
        batch_helper.connect_dataproc_cluster()
    elif action == 'stop':
        # Stop an existing DataProc cluster according to project specific configuration
        batch_helper.stop_dataproc_cluster()
    elif action == 'delete':
        # Delete an existing DataProc cluster according to project specific configuration.
        # Since we can't update certain parameters on an existing cluster, we need the option
        # to delete it
        batch_helper.delete_dataproc_cluster()


@cli.command()
@click.argument('action')
def workflow_template(action: str):
    """! Handle workflow action, either create, instantiate or delete the template
        Args:
             action: an action to perform, chosen from: {create, create_yaml, instantiate, delete}
    """
    valid_actions = {'create', 'create_yaml', 'delete', 'instantiate'}
    if action not in valid_actions:
        print(f"Can't determine action type: {action}. Please use one of: {valid_actions}")
        return

    current_folder = os.path.abspath(os.getcwd())

    os.environ["PYTHONPATH"] = current_folder
    workflow_folder = current_folder + '/batch_files/workflow/workflow.py'
    basic_command = 'python ' + workflow_folder

    print(f"Attempting to {action} a workflow template")
    subprocess.call(basic_command + ' ' + action, shell=True)


@cli.command()
@click.argument('action')
def dag(action: str):
    """! Handle DAG actions, either create, delete, or import a DAG
    Once imported, it will start as paused and will be set up for a daily activation
        Args:
             action: an action to perform, chosen from: {create, import, delete}
    """
    valid_actions = {'create', 'import', 'delete'}
    if action not in valid_actions:
        print(f"Can't determine action type: {action}. Please use one of: {valid_actions}")
        return

    current_folder = os.path.abspath(os.getcwd())

    os.environ["PYTHONPATH"] = current_folder
    dag_folder = current_folder + '/batch_files/workflow_dag/dag.py'
    basic_command = 'python ' + dag_folder

    print(f"Attempting to {action} a DAG")
    subprocess.call(basic_command + ' ' + action, shell=True)


@cli.command()
@click.argument('file_type')
def upload_batch_files(file_type: str):
    """! Upload all files needed for a batch project to the predefined bucket
        Args:
            file_type: String containing the types of files to upload (jars, stages, whl , cluster_init, or "all")
    """
    batch_helper = config_batch_helper()
    batch_helper.upload_files_to_bucket(file_type)


def config_batch_helper():
    """! Configure the batch file helper for a variety of cli actions
        Returns:
            ZIDSBatchHelper: Instantiation of the batch helper class
    """
    from dsframework.cli.batch.batch_helper_functions import ZIDSBatchHelper
    current_folder = os.path.abspath(os.getcwd())
    batch_helper_cls = None
    with open(current_folder + '/batch_files/batch_config.json') as config_file:
        config_dict = json_load(config_file)
        batch_helper_cls = ZIDSBatchHelper(config_dict)
    return batch_helper_cls


@cli.command()
@click.argument('repo_name')
def create_sonar_file(repo_name: str):
    """! Create a sonar scanner file.

    Args:
        repo_name: name of your repository.
    """
    current_folder = os.path.abspath(os.getcwd())
    data = "sonar.projectKey=dozi_" + repo_name
    create_file(os.path.join(current_folder, "sonar-scanner.properties"), data)


@cli.command()
@click.argument('csv_file_path')
@click.argument('batch_size')
def evaluate(csv_file_path, batch_size):
    """! Quick 'Initial test' to check the module.

    @verbatim
    Args:
        csv_file_path: CSV file path.
        batch_size: The size that will be taken every iteration from the CSV - in order to feed the model.
    @endverbatim
    """

    if os.path.isfile(csv_file_path):
        current_pipeline_folder = os.path.abspath(os.getcwd())
        os.environ["PYTHONPATH"] = current_pipeline_folder
        folder = current_pipeline_folder + '/tester/general_tester.py'
        subprocess.call('python ' + folder + ' ' + csv_file_path + ' ' + batch_size, shell=True)
    else:
        click.echo('file: ' + csv_file_path + ' not found')


@cli.command()
def show_documentation():
    """! Opens a browser and show the local project documentation."""
    launch_doc_webbrowser()


def launch_doc_webbrowser():
    current_pipeline_folder = os.path.abspath(os.getcwd())

    current_doc_folder = os.path.join(current_pipeline_folder, "documentation")
    if os.path.isdir(os.path.join(current_pipeline_folder, "dsframework", "documentation")):
        current_doc_folder = os.path.join(current_pipeline_folder, "dsframework", "documentation")

    project_html_path = os.path.join(current_doc_folder, 'project_docs', 'html', 'index.html')
    if os.path.isfile(project_html_path):
        webbrowser.open('file://' + project_html_path)
    else:
        click.echo('Project documentation not found, please generate documentation using: \'dsf-cli generate '
                   'documentation full\'')


ALIASES = {
    "g": generate
}


def generate_project(project_name):
    """! Generating new project (with the relevant project_name)\n
    How to generate project :
    > dsf-cli g project my-new-project
    @verbatim
    Args:
            project_name: Project name will be used in the creation of the files.
    @endverbatim
    """
    # project_name = clean_name(project_name)
    click.echo('Generating project: ' + project_name)
    create_project(project_name)


def generate_forcer(fileName):
    """! Generate new forcer - can help us, if we need forcer on an existing project """
    fileName = clean_name(fileName)
    create_exist_pipeline_file('forcer', fileName)


def generate_predictable(fileName):
    """! Generate new predictable - can help us, if we need predictable on an existing project """
    fileName = clean_name(fileName)
    create_exist_pipeline_file('predictable', fileName)


def generate_documentation(simple="full"):
    """! Document project CLI

        Usage:
            dsf-cli generate documentation full
    """
    if simple == 'full':
        click.echo('Documenting project .....')
    else:
        click.echo('Documenting project (without base).....')

    # Get folders
    current_pipeline_folder = os.path.abspath(os.getcwd())
    current_doc_folder = os.path.join(current_pipeline_folder, "documentation")
    if os.path.isdir(os.path.join(current_pipeline_folder, "dsframework", "documentation")):
        current_doc_folder = os.path.join(current_pipeline_folder, "dsframework", "documentation")

    # Validity checks
    if not os.path.isfile(os.path.join(current_doc_folder, 'project_doxyfile')):
        click.echo("Documentation folder not found, please change directory to project root folder.")
        return

    venv_path = os.getenv('VIRTUAL_ENV')

    try:
        if simple == 'full':
            os.environ["EXCLUDE_PATH"] = get_exclude_string(venv_path)
        else:
            os.environ["EXCLUDE_PATH"] = venv_path
    except Exception as e:
        click.echo(f'Switching to simple: {e}')
        os.environ["EXCLUDE_PATH"] = venv_path

    click.echo(f'Creating project documentation in {os.path.join(current_doc_folder, "project_docs")}')

    try:
        # Run doxygen as subprocess
        subprocess.call('doxygen project_doxyfile', shell=True, cwd=current_doc_folder)
    except Exception as e:
        click.echo(f'Project documentation failed!, exception: {e}')
    else:
        click.echo('Project documentation done!')
        launch_doc_webbrowser()


def get_exclude_string(venv_path):
    packages_path = site.getsitepackages()  # Get packages folder.

    # Get list of packages to exclude, except 'dsframework'
    pkg_exclude_folders = get_excluded_path_list('(?!.*dsframework)', packages_path[0], 'folders')

    # Get list of scattered file in packages folder
    pkg_exclude_files = get_excluded_path_list('(?!.*api_cli.py)', packages_path[0], 'files')

    # Get a list of files and folders in dsframework, except 'base' folder.
    pkg_exclude_dsf = get_excluded_path_list('(?!.*base)', os.path.join(packages_path[0], 'dsframework'), 'both')

    # Get venv all the rest the folders to exclude, except 'venv/lib'
    exclude_venv = get_excluded_path_list('(?!.*lib)', venv_path, 'both')

    # Combine two lists to one string: exclude_folders and exclude_folders
    exclude_full_list = pkg_exclude_folders + exclude_venv + pkg_exclude_files + pkg_exclude_dsf
    exclude_string = " \\ ".join(map(str, exclude_full_list))  # Seperated by backslash

    return exclude_string


def get_excluded_path_list(regex_str, path, get_type):

    if get_type == 'folders':
        path_list = [os.path.join(path, name) for name in os.listdir(path)
                     if os.path.isdir(os.path.join(path, name))]
    elif get_type == 'files':
        path_list = [os.path.join(path, name) for name in os.listdir(path)
                     if os.path.isfile(os.path.join(path, name))]
    else:  # 'both'
        path_list = [os.path.join(path, name) for name in os.listdir(path)]

    if regex_str != '':
        regex = re.compile(regex_str)
        return list(filter(regex.match, path_list))
    else:
        return path_list


@cli.command()
def upload_documentation():
    """! Upload project documentation CLI

        Usage:
            dsf-cli upload-documentation
    """

    # Set folders
    current_working_path = os.getcwd()
    project_name = os.path.split(current_working_path)[1]
    current_pipeline_folder = os.path.abspath(os.getcwd())
    current_doc_folder = os.path.join(current_pipeline_folder, 'documentation', 'project_docs')
    html_folder_name = 'html'

    google_gs_path = 'gs://dozi-stg-ds-apps-1-ds-apps-ds-portal/documentation/'
    project_gs_folder = f'{google_gs_path}{project_name}'
    google_storage_path = 'https://storage.cloud.google.com/dozi-stg-ds-apps-1-ds-apps-ds-portal/documentation/'
    base_tag_path = google_storage_path + project_name + '/' + html_folder_name + '/'

    # Validity checks
    if not os.path.isfile(os.path.join(current_doc_folder, html_folder_name, 'index.html')):
        click.echo('Documentation folder not found, please change directory to project root folder.')
        return

    click.echo(f'Uploading {project_name} documentation started....')

    # Create backup of current docs on google storage
    click.echo(f'Creating backup of {project_name} current documentation....')

    backup_project_name = datetime.today().strftime('%Y_%m_%d_%H_%M_%S') + '_' + project_name
    backup_destination_folder = f'{google_gs_path}backups/{backup_project_name}'

    call_gsutil_command(f'gsutil -m mv {project_gs_folder} {backup_destination_folder}',
                        project_name, current_doc_folder, 'Backup current documentation')

    # Modify project documentation, add base tag and paths
    click.echo(f'Adding <base> tag to {project_name} documentation....')
    try:
        html_list = doxygen_handler.find_files(os.path.join(current_doc_folder, html_folder_name), 'html')
        css_list = doxygen_handler.find_files(os.path.join(current_doc_folder, html_folder_name), 'css')

        doxygen_handler.modify_files(html_list, '<head>', '<head><base href=\"url_tag\">', base_tag_path)
        doxygen_handler.modify_files(css_list, "url(\'", "url(\'url_tag", base_tag_path)
        doxygen_handler.modify_files(css_list, "url(\"", "url(\"url_tag", base_tag_path)
    except Exception as e:
        click.echo(f'Adding <base> tag failed with exception: {e}')
        return
    else:
        click.echo('Adding <base> tag done!')

    # Upload current documentation
    call_gsutil_command(f'gsutil -m cp -r {os.path.join(current_doc_folder, html_folder_name)} {project_gs_folder}',
                        project_name, current_doc_folder, 'Uploading documentation')

    click.echo(f'Uploading {project_name} documentation done!!!')


def call_gsutil_command(gsutil_command, project_name, working_folder, command_desc):
    click.echo(f'{command_desc} {project_name} ....')

    try:
        subprocess.call(gsutil_command, shell=True, cwd=working_folder)
    except Exception as e:
        click.echo(f'{command_desc} failed with exception: {e}')
        return
    else:
        click.echo(f'{command_desc} done!')


def clean_name(name):
    """! Change the name from 'word1-word2' --> 'word1_word2' """
    name = name.replace('-', '_')
    return name


def create_folders(project_name):
    """! Create all needed folders for the new project creation
    @verbatim
    Args:
        project_name: Project name
    @endverbatim
    """

    global directories
    directories['main'] = project_name
    if not os.path.exists(directories['main']):
        os.mkdir(directories['main'])

    create_main_folders('config', 'main')
    create_main_folders('pipeline', 'main')
    create_main_folders('artifacts', 'pipeline')
    create_main_folders('models', 'artifacts')
    create_main_folders('vocabs', 'artifacts')
    # create_main_folders('other', 'artifacts')
    create_main_folders('preprocessor', 'pipeline')
    create_main_folders('predictables', 'pipeline')
    create_main_folders('predictors', 'pipeline')
    create_main_folders('forcers', 'pipeline')
    create_main_folders('postprocessor', 'pipeline')
    create_main_folders('schema', 'pipeline')
    create_main_folders('tester', 'main')
    create_main_folders('test_schema', 'tester')
    create_main_folders('dsp', 'tester')
    create_main_folders('trainer', 'main')
    create_main_folders('training_output', 'trainer')
    create_main_folders('pl_wrapper', 'trainer')
    create_main_folders('examples', 'pl_wrapper')
    create_main_folders('server', 'main')


def create_project(project_name):
    """! Create project from templates.
    @verbatim
    Args:
        project_name: Project name
    @endverbatim
    """

    create_folders(project_name)
    original_project_name = project_name
    project_name = clean_name(project_name)

    # Create pipeline files
    create_pipeline_file(project_name, directories['artifacts'], 'shared_artifacts')
    create_pipeline_file(project_name, directories['preprocessor'], 'preprocess')
    create_pipeline_file(project_name, directories['predictors'], 'predictor')
    create_pipeline_file(project_name, directories['forcers'], 'forcer')
    create_pipeline_file(project_name, directories['postprocessor'], 'postprocess')
    create_pipeline_file(project_name, directories['predictables'], 'predictable')
    create_pipeline_file(project_name, directories['pipeline'], 'pipeline')
    create_pipeline_file(project_name, directories['main'], 'pipeline_test', False)

    # Create schema files
    create_schema_file(project_name, directories['schema'], 'inputs')
    create_schema_file(project_name, directories['schema'], 'outputs')
    create_schema_file(project_name, directories['schema'], '__init__', False)

    # Create tester files
    create_tester_file(project_name, directories['tester'], 'general_tester', False)
    create_tester_file(project_name, directories['tester'], 'evaluator', False)
    create_schema_file(project_name, directories['test_schema'], 'test_input', False)
    create_schema_file(project_name, directories['test_schema'], 'test_output', False)

    create_dsp_file(original_project_name, directories['dsp'], 'get_model_options', False)
    create_dsp_file(original_project_name, directories['dsp'], 'get_dataset_options', False)
    create_dsp_file(original_project_name, directories['dsp'], 'add_experiment', False)

    # Create trainer files
    create_trainer_file(project_name, directories['pl_wrapper'], 'custom_dataset', False, from_folder='pl_wrapper')
    create_trainer_file(project_name, directories['pl_wrapper'], 'data_module', False, from_folder='pl_wrapper')
    create_trainer_file(project_name, directories['pl_wrapper'], 'iterable_dataset', False, from_folder='pl_wrapper')
    create_trainer_file(project_name, directories['pl_wrapper'], 'plmodel', False, from_folder='pl_wrapper')
    create_trainer_file(project_name, directories['pl_wrapper'], 'network_module', False, from_folder='pl_wrapper')
    create_trainer_file(project_name, directories['pl_wrapper'], 'trainer', False, from_folder='pl_wrapper')
    create_trainer_file(project_name, directories['examples'], 'example1_mnist', False,
                        from_folder=os.path.join('pl_wrapper', 'examples'))
    create_trainer_file(project_name, directories['examples'], 'example2_wine_quality', False,
                        from_folder=os.path.join('pl_wrapper', 'examples'))
    create_trainer_file(project_name, directories['trainer'], 'comparison', False)
    create_trainer_file(project_name, directories['trainer'], 'config', False)
    create_trainer_file(project_name, directories['trainer'], 'data', False)
    create_trainer_file(project_name, directories['trainer'], 'model', False)
    create_trainer_file(project_name, directories['trainer'], 'test', False)
    create_trainer_file(project_name, directories['trainer'], 'train', False)
    create_trainer_file(project_name, directories['trainer'], 'trainer_main', False)
    source_path = os.path.join(__location__, 'dsframework/cli/trainer/requirements_trainer.txt')
    shutil.copyfile(source_path, directories['trainer'] + '/requirements_trainer.txt')

    # Create server files
    create_server_file(project_name, directories['server'], 'main', False)
    create_server_file(project_name, directories['server'], 'test_server_post', False)
    create_server_file(project_name, directories['server'], 'pool', False)
    create_server_file(project_name, directories['server'], 'token_generator', False)
    create_server_file(project_name, directories['server'], '__init__', False)

    create_project_config_json()
    create_project_gitignore()
    create_server_config_json()
    copy_deploy_files(directories['main'])
    copy_cloud_eval_files(directories['main'])
    copy_batch_files(directories['main'])
    copy_documentation_files(os.path.join(directories['main'], 'documentation'))
    update_doc_configuration(project_name, os.path.join(directories['main'], 'documentation'))

    change_to_project_dir()
    run_dvc_init()
    run_git_init()


def create_main_folders(targetDir, baseDir):
    """! Create folder if not already exists
    @verbatim
    Args:
        targetDir: the new folder to create
        baseDir  : where we want to create the new folder.
    @endverbatim
    """

    global directories
    directories[targetDir] = directories[baseDir] + '/' + targetDir
    if not os.path.exists(directories[targetDir]):
        os.mkdir(directories[targetDir])


def create_pipeline_file(project_name, folder, pipelineType, createInitFile=True):
    """! Takes the template file and replace it with new one with project_name
    @verbatim
    Args:
        project_name : The project name
        folder       : The folder we want to create the new file.
        pipelineType : File type in the pipeline.
        createInitFile(=True): Create init file - if doesnt exist.
    @endverbatim
    """

    data = read_template_file(pipelineType)
    replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile)


def create_schema_file(project_name, folder, pipelineType, createInitFile=True):
    """! Create schema files (read the template and replace it )
    @verbatim
    Args:
        project_name : The project name
        folder       : The folder we want to create the new file.
        pipelineType : File type. TODO change pipelineType name
        createInitFile(=True): Create init file - if doesnt exist.
    @endverbatim
    """
    data = read_template_file('schema/' + pipelineType)
    replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile)


def create_server_file(project_name, folder, pipelineType, createInitFile=True):
    """! Create server file (read the template and use him to create the new file)
    @verbatim
    Args:
        project_name : The project name
        folder       : The folder we want to create the new file.
        pipelineType : File type. TODO change pipelineType name
        createInitFile(=True): Create init file - if doesnt exist.
    @endverbatim
    """
    data = read_template_file('tester/server/' + pipelineType)
    replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile)


def create_tester_file(project_name, folder, test_file_type, createInitFile=True):
    """! Create tester file (read the template and use him to create the new file)
    @verbatim
    Args:
        project_name   : The project name.
        folder         : The folder we want to create the new file.
        test_file_type : File type.
        createInitFile(=True): Create init file - if doesnt exist.
    @endverbatim
    """
    data = read_template_file('tester/' + test_file_type)
    replace_in_template_and_create_file(project_name, folder, test_file_type, data, createInitFile)


def create_trainer_file(project_name, to_folder, test_file_type, createInitFile=True, from_folder=''):
    """! Create trainer file (read the template and use him to create the new file)
    @verbatim
    Args:
        project_name   : The project name.
        folder         : The folder we want to create the new file.
        test_file_type : File type.
        createInitFile(=True): Create init file - if it doesn't exist.
    @endverbatim
    """

    # data = read_template_file('trainer/' + test_file_type)
    data = read_template_file(os.path.join('trainer', from_folder, test_file_type))
    replace_in_template_and_create_file(project_name, to_folder, test_file_type, data, createInitFile)


def create_dsp_file(project_name, folder, test_file_type, createInitFile=True):
    """! Create DSP file (read the template and use him to create the new file)
    @verbatim
        Args:
            project_name   : The project name.
            folder         : The folder we want to create the new file.
            test_file_type : File type.
            createInitFile(=True): Create init file - if doesnt exist.
    @endverbatim
    """
    data = read_template_file('tester/dsp/' + test_file_type)
    replace_in_template_and_create_file(project_name, folder, test_file_type, data, createInitFile)


def create_exist_pipeline_file(type, fileName):
    """! Create a new component in the pipeline and integrate it (for example forcer or predictable)
    @verbatim
    Args:
        type      : New component type (forcer or predictable)
        fileName  : New file name.
    @endverbatim
    """

    pipelineType = type + 's'
    folder = 'pipeline/' + pipelineType
    if os.path.exists(folder):
        data = read_template_file(type)
        fileNameNoUnderscore = to_capitalize_no_underscore(fileName)
        className = fileNameNoUnderscore + type.capitalize()
        currentPipelineFolder = os.path.basename(os.getcwd())
        currentDir = folder.replace('/', '.')

        data = data.replace('generatedClass', className)

        new_file = folder + "/" + fileName + ".py"
        current_init_file = folder + "/__init__.py"
        new_init_export = "from " + '.' + fileName + " import " + className

        create_file(new_file, data)
        create_init_file(current_init_file, new_init_export)
        inject_to_pipeline(fileName, type, className, new_init_export)
    else:
        print('please create a project and go to project location first')
    pass


def create_tester_env():
    """! This function will create a testing environment, if one does not already exist in the project.
    It will create the testing, evaluation, and relevant schema files.
    """

    project_name = os.getcwd().split('/')[-1]
    dst_folder = 'tester/'
    schema_dst_folder = 'tester/test_schema/'
    dsp_dst_folder = 'tester/dsp/'
    tester_file_list = ["tester/general_tester",
                        "tester/evaluator",
                        "schema/test_input",
                        "schema/test_output",
                        "tester/dsp/add_experiment",
                        "tester/dsp/get_dataset_options",
                        "tester/dsp/get_model_options"]

    if not os.path.exists(dst_folder):
        os.mkdir(dst_folder)
        os.mkdir(schema_dst_folder)
        os.mkdir(dsp_dst_folder)

        for filename_with_path in tester_file_list:
            file_only = filename_with_path.split('/')[-1]
            parent_folder = filename_with_path.split('/')[-2]
            if parent_folder == 'schema':
                dst_folder = schema_dst_folder
            elif parent_folder == 'dsp':
                dst_folder = dsp_dst_folder

            data = read_template_file(filename_with_path)

            clean_project_name = clean_existing_project_name(project_name)
            filename_no_underscore = to_capitalize_no_underscore(file_only)

            class_name = clean_project_name + filename_no_underscore

            data = data.replace('generatedClass', class_name)
            data = data.replace('generatedProjectName', clean_project_name)

            new_file = dst_folder + "/" + file_only + ".py"
            create_file(new_file, data)

            # Test schema does not use generated class name for now, so this code should be modified before uncommenting
            # current_init_file = dst_folder + "/" + "__init__.py"
            # new_init_export = "from " + dst_folder.replace('/', '.') + "." + file_only + " import " + class_name
            # create_init_file(current_init_file, new_init_export)
    else:
        print('Tester folder Already exists. If you wish to recreate, please erase folder and retry')
    pass


def read_template_file(filename):
    """! Read \<filename\>_template.py  from the cli dir """
    with open(os.path.join(__location__, 'dsframework/cli/' + filename + '_template.py'), 'r') as file:
        data = file.read()
        return data


def read_file(file_path):
    with open(file_path, 'r') as file:
        data = file.read()
        return data


def write_to_file(file_path, data):
    with open(file_path, 'w') as file:
        file.write(data)
        file.close()


def replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile):
    """! Get & Replace the template file and create new one with project_name
    @verbatim
    Args:
        project_name : The project name
        folder       : The folder we want to create the new file.
        pipelineType : Which File.
        data         : The template file.
        createInitFile(=True): Create init file - if doesnt exist.
    @endverbatim
    """

    pipeline_type_no_underscore = to_capitalize_no_underscore(pipelineType)
    project_name_no_underscore = to_capitalize_no_underscore(project_name)
    className = project_name_no_underscore + pipeline_type_no_underscore
    classNameForBaseObject = project_name_no_underscore + pipeline_type_no_underscore
    currentDir = folder.replace('/', '.')
    currentDirNoMainDir = folder.replace(directories['main'] + '/', '').replace('/', '.')
    currentBaseDir = directories['main'].replace('/', '.')
    currentPipelineDir = directories['pipeline'].replace('/', '.')

    data = data.replace('generatedClassName', classNameForBaseObject)
    data = data.replace('generatedClass', className)
    data = data.replace('original_project_name', project_name)
    data = data.replace('generatedProjectName', project_name_no_underscore)
    data = data.replace('generatedDirectory', currentDir)
    data = data.replace('generatedBaseDir', currentBaseDir)
    data = data.replace('generatedPipelineDir', currentPipelineDir)

    new_file = folder + "/" + pipelineType + ".py"
    create_file(new_file, data)

    if createInitFile:
        new_init_file = folder + "/__init__.py"
        new_init_export = "from " + '.' + pipelineType + " import " + className
        create_init_file(new_init_file, new_init_export)


def create_predictable_file(fileName):
    pass


def create_file(new_file_path, data):
    """! Create new file and write the data file into it
    @verbatim
    Args:
        new_file_path : New file name path.
        data          : The content of the new file.
    @endverbatim
    """

    if not os.path.exists(new_file_path):
        f = open(new_file_path, "w")
        f.write(data)
        f.close()


def create_init_file(init_path, init_export):
    """! if init file doesnt exist - create it and add the export line
         if init file exist - check if the init_export line appears in the init file : if not add it.

    @verbatim
    Args:
        init_path   : Init file.
        init_export : Import line to add if needed.
    @endverbatim
    """

    if not os.path.exists(init_path):
        f = open(init_path, "w")
        f.write(init_export)
        f.close()
    else:
        f = open(init_path, 'r+')
        data = f.read()
        if init_export not in data:
            if len(data) and not data.endswith('\n'):
                f.write('\n')
            f.write(init_export)
            f.close()


def save_file(file_name_path, data):
    f = open(file_name_path, "w")
    f.write(data)
    f.close()


def inject_to_pipeline(fileName, type, className, new_init_export):
    """! Building the component in the pipeline file
    @verbatim
    Args:
        fileName       : New file name.
        type           : New component type (forcer or predictable)
        className      : New component class.
        new_init_export: Create init file - if doesnt exist.
    @endverbatim
    """

    file_path = 'pipeline/pipeline.py'
    if os.path.exists(file_path):
        data = read_file(file_path)
        new_tab_line = '\n'
        data_changed = False
        last_index_of_import = -1
        first_index_of_add_component = -1
        index_of_class = re.search(r'class[^\n]*', data)

        # finding current indent config
        index_of_build_pipeline = re.search(r'def build_pipeline[^\n]*', data)
        index_of_build_preprocessor = re.search(r'self.preprocessor =[^\n]*', data)
        index_of_build_postprocessor = re.search(r'self.postprocessor =[^\n]*', data)
        index_of_build_pipeline_row = re.search(r'.*def build_pipeline[^\n]*', data)
        if index_of_build_pipeline:
            build_pipeline_indent = (index_of_build_pipeline.start() - index_of_build_pipeline_row.start()) * 2
            new_tab_line = new_tab_line.ljust(build_pipeline_indent + 1)

        attr = fileName + type.title()
        new_component_line = 'self.' + attr + ' = ' + className + '()'
        add_component_line = 'self.add_component(self.' + attr + ')'

        # finding imports and add components indexes
        last_index_of_add_component = -1
        all_from_import = [i.end() for i in re.finditer(r'from[^\n]*', data)]
        all_add_components = [[i.start(), i.end()] for i in re.finditer(r'self.add_component[^\n]*', data)]
        if len(all_from_import):
            last_index_of_import = all_from_import[-1]
        if len(all_add_components):
            first_index_of_add_component = all_add_components[0][0]
            last_index_of_add_component = all_add_components[-1][-1]
        # finding imports and add components indexes


        index_to_add = 0

        # add import to end of imports or to top of file
        if last_index_of_import > -1 and new_init_export not in data:
            s = '\n' + new_init_export
            index_to_add += len(s)
            data = data[:last_index_of_import] + s + data[last_index_of_import:]
            data_changed = True
        elif index_of_class and new_init_export not in data:
            s = new_init_export + '\n\n'
            index_to_add += len(s)
            data = data[:index_of_class.start()] + s + data[index_of_class.start():]
            data_changed = True

        # check if build_pipeline exist but with no components yet
        if first_index_of_add_component == -1 and last_index_of_add_component == -1 and index_of_build_pipeline:
            s = new_tab_line
            current_end = index_of_build_pipeline.end()
            if index_of_build_preprocessor:
                current_end = index_of_build_preprocessor.end()
            if index_of_build_postprocessor:
                current_end = index_of_build_postprocessor.end()
            index = current_end + index_to_add
            index_to_add += len(s)
            data = data[:index] + s + data[index:]
            data_changed = True
            first_index_of_add_component = current_end
            last_index_of_add_component = current_end

        # adding new component line
        if first_index_of_add_component > -1 and new_component_line not in data:
            first_index_of_add_component += index_to_add
            if len(all_add_components):
                new_component_line = new_component_line + new_tab_line
            index_to_add += len(new_component_line)
            data = data[:first_index_of_add_component] + new_component_line + data[first_index_of_add_component:]
            data_changed = True

        #adding add_component line
        if last_index_of_add_component > -1 and add_component_line not in data:
            last_index_of_add_component += index_to_add
            add_component_line = new_tab_line + add_component_line
            index_to_add += len(add_component_line)
            data = data[:last_index_of_add_component] + add_component_line + data[last_index_of_add_component:]
            data_changed = True

        if data_changed:
            write_to_file(file_path, data)


def create_project_config_yaml():
    """! Creates /pipeline/config.yaml from dsframework/cli/config.yaml """
    with open(os.path.join(__location__, 'dsframework/cli/config.yaml'), 'r') as file:
        data = file.read()
        data = data.replace('generatedDirectory', directories['main'])
        new_file = directories['main'] + '/pipeline/config.yaml'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def create_project_config_json():
    """! Creates /config/config.json from dsframework/cli/config.json """
    with open(os.path.join(__location__, 'dsframework/cli/config.json'), 'r') as file:
        data = file.read()
        data = data.replace('generatedDirectory', directories['main'])
        new_file = directories['main'] + '/config/config.json'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def create_project_gitignore():
    """! Creates /.gitignore from dsframework/cli/.gitignore
         Which means all files that appear in the project directory and should not be uploaded to the git repository.
    """
    with open(os.path.join(__location__, 'dsframework/cli/.gitignore'), 'r') as file:
        data = file.read()
        new_file = directories['main'] + '/.gitignore'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def create_server_config_json():
    """! Creates server/cors_allowed_origins from dsframework/cli/cors_allowed_origins   """
    with open(os.path.join(__location__, 'dsframework/cli/cors_allowed_origins.json'), 'r') as file:
        data = file.read()
        data = data.replace('generatedDirectory', directories['main'])
        new_file = directories['server'] + '/cors_allowed_origins.json'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def change_to_project_dir():
    """! Change the OS dir path """
    os.chdir(directories['main'])


def run_dvc_init():
    """! Running dvc init and doing the following:
    1. Defines where all the dvc will be uploaded (DVC_BUCKET=)
    2. dvc init    - generate a local repository
    3. dvc config
    4. dvc remote add - connect my local repo -> remote
    5. dvc remote modify
    """
    dir_path = os.getcwd()
    if not os.path.isdir(dir_path + '/.dvc'):
        command = 'dvc_init.sh'
        if not isWindows:
            command = './' + command
        subprocess.call(command, shell=True)


def run_git_init():
    """! Using git_init.sh to create a new project on git
        https://git.zoominfo.com/dozi/{name-your-service}
    """
    dir_path = os.getcwd()
    if not os.path.isdir(dir_path + '/.git'):
        command = 'git_init.sh'
        if not isWindows:
            command = './' + command
        subprocess.call(command, shell=True)


def copy_deploy_files(main_dir):
    """! Copy deploy files (from dsframework/cli/tester/deploy_files """
    currentPipelineFolder = os.path.basename(os.getcwd())
    if main_dir:
        currentPipelineFolder = main_dir
    dir = os.path.join(__location__, 'dsframework/cli/tester/deploy_files/')
    listOfFiles = list()
    for (dirpath, dirname, filenames) in os.walk(dir):
        dirpath = dirpath.replace(dir, '')
        listOfFiles += [os.path.join(dirpath, file) for file in filenames]
        for file_path in listOfFiles:
            with open(os.path.join(dir, file_path), 'r', encoding="utf8") as file:
                try:
                    data = file.read()
                    data = data.replace('{name-your-service}', currentPipelineFolder)
                    data = data.replace('{name-your-artifacts}', currentPipelineFolder)
                    dirToCreate = ''
                    if main_dir:
                        dirToCreate = main_dir + '/'
                    if not os.path.exists(dirToCreate + 'deploy'):
                        os.makedirs(dirToCreate + 'deploy')
                    if not os.path.exists(dirToCreate + file_path):
                        f = open(dirToCreate + file_path, "w")
                        f.write(data)
                        f.close()
                        # Make sure that .sh files can be executed
                        if (dirToCreate + file_path).endswith('.sh'):
                            os.chmod(dirToCreate + file_path, 0o744)
                except Exception as ex:
                    pass
    # for item in os.listdir(dir):
    #     s = os.path.join(dir, item)
    #     d = os.path.join(main_dir, item)
    #     if os.path.isdir(s):
    #         shutil.copytree(s, d, False, None)
    #     else:
    #         shutil.copy2(s, d)


def copy_cloud_eval_files(main_dir):
    """! Copy cloud evaluation files """
    currentPipelineFolder = os.path.basename(os.getcwd())
    if main_dir:
        currentPipelineFolder = main_dir
    service_folder = str(currentPipelineFolder)
    service_name = service_folder.split('/')[-1]
    project_name_no_underscore = clean_existing_project_name(service_name)
    cloud_eval_dir = os.path.join(__location__, 'dsframework/cli/tester/cloud_eval/')
    listOfFiles = list()
    conflicted_files = []
    for (dirpath, dirname, filenames) in os.walk(cloud_eval_dir):
        if "__pycache__" in dirpath:
            continue
        dirpath = dirpath.replace(cloud_eval_dir, '')
        listOfFiles += [os.path.join(dirpath, file) for file in filenames]
        for file_path in listOfFiles:
            framework_source_file = os.path.join(cloud_eval_dir, file_path)
            with open(framework_source_file, 'r', encoding="utf8") as file:
                try:
                    data = file.read()
                    data = data.replace('{name-your-service}', service_name)
                    data = data.replace('{name-your-artifacts}', service_name)
                    data = data.replace('generatedProjectName', project_name_no_underscore)
                    dirToCreate = ''
                    if main_dir:
                        dirToCreate = main_dir + '/cloud_eval/'
                    if not os.path.exists(dirToCreate):
                        os.makedirs(dirToCreate)
                    if "#cli_file_dest=[" in data:
                        start_ix = data.index("#cli_file_dest=[") + len("#cli_file_dest=[")
                        end_ix = data.index("]", start_ix)
                        destination_dir = data[start_ix:end_ix]
                        data = data[:data.index("#cli_file_dest=[")] + data[end_ix+1:]
                        dirToCreate = main_dir + '/' + destination_dir

                    if not os.path.exists(dirToCreate + file_path):
                        f = open(dirToCreate + file_path, "w")
                        f.write(data)
                        f.close()
                        # Make sure that .sh files can be executed
                        if (dirToCreate + file_path).endswith('.sh'):
                            os.chmod(dirToCreate + file_path, 0o744)
                    else:
                        with open(dirToCreate + file_path, "r") as f:
                            existing_data = f.read()
                        if existing_data != data:
                            conflicted_file = f"{dirToCreate + file_path}"
                            conflicted_files.append((conflicted_file, str(framework_source_file)))

                except Exception as ex:
                    pass

    if conflicted_files:
        print("The following files already exists, so not overwriting them. "
              "However, notice it is different from the framework version in")
        for conflicted_file in conflicted_files:
            print(f"Existing: [{conflicted_file[0]}] \t\t Framework source file [{conflicted_file[1]}]")


def copy_batch_files(main_dir):
    """! Generate batch files needed for batch projects. Creates 3 stages:
    1. Get Data
    2. Run Pipeline
    3. Set Data

    Along with a matching workflow file, and a configuration file

        Args:
            main_dir: A string describing the main working directory, on existing projects, it's empty.
    """

    current_folder = os.path.basename(os.getcwd())
    if main_dir != '':
        current_folder = main_dir
    service_name = current_folder.split('/')[-1]
    project_name_no_underscore = clean_existing_project_name(service_name)
    source_dir = os.path.join(__location__, 'dsframework/cli/tester/batch_files/')

    for (dir_path, dir_name, filenames) in os.walk(source_dir):
        if "__pycache__" in dir_path:
            continue

        dir_path = dir_path.replace(source_dir, '')
        file_list = [os.path.join(dir_path, file) for file in filenames]
        for file_path in file_list:
            with open(os.path.join(source_dir, file_path), 'r', encoding="utf8") as file:
                try:
                    data = file.read()
                    data = data.replace('{name-your-service}', service_name)
                    data = data.replace('generatedProjectName', project_name_no_underscore)

                    if main_dir == '':
                        target_dir = 'batch_files/'
                        dir_to_create = target_dir
                    else:
                        target_dir = current_folder + '/batch_files/'
                        dir_to_create = target_dir

                    if dir_path != "":
                        dir_to_create = target_dir + dir_path + '/'

                    if not os.path.exists(dir_to_create):
                        os.makedirs(dir_to_create)
                    if not os.path.exists(target_dir + file_path):
                        f = open(target_dir + file_path, "w")
                        f.write(data)
                        f.close()
                except Exception as ex:
                    if file_path.rsplit('.', 1)[-1] == 'jar':
                        shutil.copy(os.path.join(source_dir, file_path), f'{dir_to_create}')
                    else:
                        print(f'Error when trying to open: {file_path}: {ex}')
                    pass


def copy_trainer_files():
    """! This function will create a trainer environment."""

    project_name = os.getcwd().split('/')[-1]
    trainer_main_folder = 'trainer/'
    trainer_datasets_folder = 'trainer/datasets/'
    trainer_datasets_versions_folder = 'trainer/datasets/my_dataset_name/' + datetime.today().strftime('%Y%m%d_v1') + '/'
    trainer_examples_folder = 'trainer/examples/'

    trainer_file_list = ["trainer/custom_dataset",
                         "trainer/data_module",
                         "trainer/iterable_dataset",
                         "trainer/model",
                         "trainer/network_module",
                         "trainer/trainer",
                         "trainer/examples/example1_mnist",
                         "trainer/examples/example2_wine_quality"]

    if not os.path.exists(trainer_main_folder):
        os.mkdir(trainer_main_folder)

    if not os.path.exists(trainer_datasets_folder):
        os.makedirs(trainer_datasets_versions_folder)

    if not os.path.exists(trainer_examples_folder):
        os.mkdir(trainer_examples_folder)

    for filename_path in trainer_file_list:
        file_only = filename_path.split('/')[-1]
        parent_folder = filename_path.split('/')[-2]

        dst_folder = ''
        if parent_folder == 'datasets':
            dst_folder = trainer_datasets_folder
        elif parent_folder == 'examples':
            dst_folder = trainer_examples_folder

        data = read_template_file(filename_path)

        clean_project_name = clean_existing_project_name(project_name)
        filename_no_underscore = to_capitalize_no_underscore(file_only)

        class_name = clean_project_name + filename_no_underscore

        data = data.replace('generatedClass', class_name)
        data = data.replace('generatedProjectName', clean_project_name)

        new_file = trainer_main_folder + file_only + ".py"
        if dst_folder != '':
            new_file = dst_folder + file_only + ".py"

        if not os.path.isfile(new_file):
            create_file(new_file, data)
        else:
            print(f'File \'{new_file}\' exists, skipping.')

    source_path = os.path.join(__location__, 'dsframework/cli/', 'trainer/requirements_trainer.txt')
    shutil.copyfile(source_path, trainer_main_folder + '/requirements_trainer.txt')

    print("Creating trainer environment done !")


def copy_documentation_files(target_dir):
    """! Copy documentation folder to newly created project.

    Args:
        target_dir: Newly created project location
    """
    source_dir = os.path.join(__location__, 'dsframework/documentation/')
    shutil.copytree(source_dir, target_dir, ignore=shutil.ignore_patterns('dsf_doxy*'))


def update_doc_configuration(project_name, target_dir):
    """! Update project doxygen configuration file, adding project name.

    Args:
        project_name: Newly created project name
        target_dir: Configuration file location.
    """
    project_name_no_underscore = to_capitalize_no_underscore(project_name)
    source_dir = os.path.join(target_dir, 'project_doxyfile')
    data = read_file(source_dir)
    data = data.replace('generatedProjectName', project_name_no_underscore)
    save_file(source_dir, data)


@cli.command()
def install_trainer_packages():
    """! Install trainer packages."""
    command = 'pip install -r trainer/requirements_trainer.txt'
    subprocess.call(command, shell=True)


@cli.command()
def install_documentation_requirements():
    """! Install documentation doxygen."""

    doxygen_found = True

    try:  # Check for brew installation
        command = 'brew -v'
        res = subprocess.check_output(command, shell=True)
        brew_found = 'Homebrew' in str(res)

        if brew_found:
            try:  # Check for doxygen installation (if brew found)
                command = 'brew ls --versions doxygen'
                res = subprocess.check_output(command, shell=True)
                doxygen_found = 'doxygen' in str(res)

                if doxygen_found:
                    print('Doxygen already installed.')

            except subprocess.CalledProcessError as e1:
                doxygen_found = False

    except subprocess.CalledProcessError as e2:
        brew_found = False

    if not doxygen_found or not brew_found:
        installation_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'dsframework', 'documentation',
                                         'install_doxygen.sh')
        if not os.path.exists(installation_file):
            print('Installation script not found install_doxygen.sh')
            return
        is_mac_linux = (platform.system() == 'Darwin' or platform.system() == 'Linux')

        if is_mac_linux:
            subprocess.call(f'chmod +x {installation_file}', shell=True)

        subprocess.call(installation_file, shell=True)


def to_capitalize_no_underscore(text):
    """! Capitalize each word with no underscore
     for example : Moving from test_name -> TestName"""
    return ''.join(elem.capitalize() for elem in text.split('_'))


def clean_existing_project_name(text):
    no_underscore_text = to_capitalize_no_underscore(text)
    return ''.join(elem.capitalize() for elem in no_underscore_text.split('-'))


def let_user_pick(options):
    """! Let the user pick a number from couple of options
    @verbatim
    Args:
        options  : couple of options for user decision.
    @endverbatim
    """
    print("Please choose:")
    for idx, element in enumerate(options):
        print("{}) {}".format(idx+1, element))
    i = input("Enter number: ")
    try:
        if 0 < int(i) <= len(options):
            return int(i)-1
    except:
        pass
    return None


@cli.command()
def deploy_service_to_gae():
    """! Run deployment scripts in order to deploy the current project to GAE.
   """
    # Setup paths to script files
    current_folder = os.path.abspath(os.getcwd())
    setup_file_path = f"{current_folder}/deploy/setup_deploy_params.sh"
    framework_folder = os.path.dirname(os.path.abspath(__file__))

    # Check that we have a setup file. If not, copy it.
    if not os.path.exists(setup_file_path):
        print(f"file {setup_file_path} doesn't exist. Copying. Please fill relevant parameters and run again")
        shutil.copyfile(f"{framework_folder}/dsframework/cli/tester/deploy_files/deploy/setup_deploy_params.sh",
                        setup_file_path)
        return

    # Prepare commands
    setup_command = f"source {setup_file_path}"
    build_command = f"{framework_folder}/dsframework/base/deploy_scripts/build_and_register_image_in_devland.sh"
    deploy_command = f"{framework_folder}/dsframework/base/deploy_scripts/gae/deploy-gae-devland.sh"
    full_command = f"{setup_command};{build_command};{deploy_command}"

    # Execute
    subprocess.call(full_command, shell=True)
    #subprocess.call(setup_command, shell=True)
    #subprocess.call(build_command, shell=True)
    #subprocess.call(deploy_command, shell=True)


if __name__ == '__main__':
    cli(prog_name='cli')



from dsframework.base.trainer.model import ZIDSTrainerModel
from trainer.pl_wrapper.plmodel import generatedProjectNamePlmodel

##
# @file
# @brief generatedClass class, defines model class.
class generatedClass(ZIDSTrainerModel):
    """Model class - defines the network, loss function etc."""
    model = None
    config = None
    metrics_config = None

    def __init__(self, config, metrics_config):
        super().__init__(config, metrics_config)
        self.metrics_config = metrics_config

    def define_model(self):
        """! Implement model class in the format you use in the fit/train method.
        Create a new model or load from pretrained model.

        Needs to be implemented.

        Example (pytorch lightning):
            self.model = generatedProjectNamePlmodel(
                loss_function=self.config['loss_function'],
                num_classes=self.config['num_classes'],
                num_classes_conf_matrix=self.config['num_classes_confusion_matrix'],
                learn_rate=self.config['learning_rate'],
                nnetwork=self.config['nnetwork'],  # Create a new model or load from pretrained model.
                metrics=self.metrics_config
            )
        """
        if self.model is None:
            raise Exception('Implement model in model.py')

    def get_model(self):
        """Returns model class"""
        return self.model



class ZIDSTrainerModel:

    model = None
    config = None
    metrics_config = None

    def __init__(self, config, metrics_config):
        self.config = config
        self.metrics_config = metrics_config
        self.define_model()

    def define_model(self):
        self.model = None

    def get_model(self):
        return self.model



from os.path import exists
from typing import Optional, Union, Any
import csv
from dsframework.base.trainer.pl_wrapper import *


class ZIDSDataModule(pl.LightningDataModule):
    """! This class inherits from LightningDataModule"""

    def __init__(self, batch_size, seed=42, num_workers=1):
        """! Define required parameters here."""
        super().__init__()
        self.dataset = None
        self.batch_size = batch_size
        self.seed = seed
        self.num_workers = num_workers
        self.dataset = None
        self.ds_train = None
        self.ds_val = None
        self.ds_test = None
        self.ds_predict = None

    def prepare_data(self):
        """! Define steps that should be done on only one GPU, like getting data."""
        pass

    def setup(self, stage: Optional[str] = None) -> None:
        """! Make assignments here (val/train/test split)
        called on every process in DDP
        Define steps that should be done on
        every GPU, like splitting data, applying
        transform etc.


        Example:
            @code
            # Assign train/val datasets for use in dataloaders:
            if stage in (None, "fit"):
                mnist_full = MNIST(self.data_dir, train=True, transform=self.transform)
                self.ds_train, self.ds_val = random_split(mnist_full, [55000, 5000])

            # Assign test dataset for use in dataloader(s):
            if stage in (None, "test"):
                self.ds_test = MNIST(self.data_dir, train=False, transform=self.transform)

            @endcode
        """
        pass

    def train_dataloader(self) -> TRAIN_DATALOADERS:
        """! Return DataLoader for Training Data here

        Example:
            @code
            train_split = Dataset(...)
            return DataLoader(train_split)
            @endcode
        """
        return DataLoader(self.ds_train, batch_size=self.batch_size, num_workers=self.num_workers)

    def val_dataloader(self) -> EVAL_DATALOADERS:
        """! Return DataLoader for data validation.

        Example:
            @code
            val_split = Dataset(...)
            return DataLoader(val_split)
            @endcode
        """
        return DataLoader(self.ds_val, batch_size=self.batch_size, num_workers=self.num_workers)

    def test_dataloader(self) -> EVAL_DATALOADERS:
        """! Return DataLoader for data testing.

        Example:
            @code
            val_split = Dataset(...)
            return DataLoader(test_split)
            @endcode
        """
        return DataLoader(self.ds_test, batch_size=self.batch_size, num_workers=self.num_workers)

    def predict_dataloader(self) -> EVAL_DATALOADERS:
        return DataLoader(self.ds_predict, batch_size=self.batch_size)

    def teardown(self, stage: Optional[str] = None) -> None:
        """! Called at the end of fit(train + validate), validate, test, or predict.
            called on every process in DDP
        """
        pass

    def load_dataset(self, file_path, file_type, remove_header=True):
        """! Loads an external dataset, NOT FINISHED YET."""
        if file_type == 'csv':
            with open(file_path) as csv_file:
                self.dataset = list(csv.reader(csv_file))

        return self.dataset

    def parse_dataset(self, remove_header) -> Union[Any, Any]:
        """! Parse your data here, it needs to return x, y.
            Returns:
                x: dataset
                y: target

        Example:
            @code{.py}
            if remove_header:
                self.dataset = self.dataset[1:]

            x = np.array(self.dataset, dtype=np.float32)[:, :-1]
            y = np.array(self.dataset, dtype=np.float32)[:, -1].reshape(-1, 1)

            return x, y
            @endcode
        """
        raise Exception("Not implemented exception (parse_dataset).")




"""! @brief Run pipeline stage, which will run the project's pipeline on the retrieved data."""
import sys
import pandas as pd
from typing import Any
from json import loads as json_loads
from pyspark.sql.types import *
from typing import Iterator

from dsframework.base.batch.stage_base import ZIDS_Stage


##
# @file
# @brief Stage main class, implements ZIDS_Stage base class.
class RunPipelineStage(ZIDS_Stage):
    """! Stage class

    Implement a stage that will later be converted to an executable job in a specific workflow.
    """

    def __init__(self, stage_config):
        """! The Stage class (RunPipelineStage) initializer.
        Base class will load basic configuration parameters, additional fields should be added here.

            Args:
                stage_config : Configuration dictionary, loaded from configuration file.
        """

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(stage_config)
        self.udf_schema = StructType([StructField("id", IntegerType()),
                                      StructField("url", StringType()),
                                      StructField("res", StringType())])

    def get_stage_name(self):
        """! Get the stage name

            Returns:
                A string, containing the stage's name
        """
        return self.__class__.__name__

    @staticmethod
    def map_in_pandas(partition: Iterator[pd.DataFrame]):
        """! Executes the project main pipeline

                Args:
                    partition: Iterator[pd.DataFrame], we get an iterator to a Pandas dataframe, on which
                    we can perform actions, such as running the project pipeline and writing the output
        """
        def test():
            from pipeline.pipeline import generatedProjectNamePipeline
            p = generatedProjectNamePipeline()
            pipeline_output = p.execute()
            return pipeline_output

        for pdf in partition:
            pdf_output = test()
            print(f'output: {pdf_output}')
            pdf.at[0, 'res'] = pdf_output.out_text
            yield pdf[pdf.id == 2]

    def main(self, **kwargs: Any):
        """! Executes the main functionality of the stage.

            Args:
                **kwargs : Whatever is needed for the stage to run properly.
        """
        spark_session = self.load_spark()
        df = spark_session.createDataFrame([(1, 'http://aaa', "I"),
                                            (2, 'http://bbb', "am"),
                                            (3, 'http://ccc', "testing")], ("id", "url", "res"))
        df.show(3)
        out_df = df.mapInPandas(func=self.map_in_pandas, schema=self.udf_schema)
        out_df.show(1)


if __name__ == "__main__":
    """! Executes the stage by instantiating it and calling the main function.
    Set up argument condition according to the usage of the written stage

        Args:
            System argument 1 - Configuration file
            System argument 2 - Start date
    """
    if sys.argv and len(sys.argv) > 1:
        config = json_loads(sys.argv[1])
        stage = RunPipelineStage(config)
        try:
            start_date = sys.argv[2]
            end_date = sys.argv[3]
            params = sys.argv[4]
            stage.update_stage_params(start_date, end_date, params)
            stage.main()
        except Exception as e:
            raise Exception(f" Stage failed with error: {e}")
    else:
        raise Exception(f"Stage configuration not provided, Can't run stage")





from dsframework.base.trainer.data import ZIDSTrainerData
from trainer.pl_wrapper.data_module import generatedProjectNameDataModule

##
# @file
# @brief generatedClass class, Datasets and functionality related to datasets.
class generatedClass(ZIDSTrainerData):
    """generatedClass, holds datasets and functionality related to it."""
    dataset_train = None
    dataset_validation = None
    dataset_test = None
    data_module = None

    def __init__(self, dataset_train=None, dataset_validation=None, dataset_test=None, save_path=''):
        """generatedClass class initializer."""
        super().__init__(dataset_train, dataset_validation, dataset_test, save_path)

    def create_data_module(self):
        """! Creates data module in the format you use in the fit/train method. Called from super().__init__().

        Needs to be implemented.

        Example (pytorch lightning):
            self.data_module = generatedProjectNameDataModule(
            dataset=self.dataset_train,
            valid_set=self.dataset_validation,
            test_set=self.dataset_test)

            return self.data_module
        """
        if self.data_module is None:
            raise Exception('Implement data module in data.py -> create_data_module()')

    def get_data_module(self):
        """! Returns data module in the format you use in the fit/train method.

        Returns:
            data_module

        """
        if self.data_module is None:
            raise Exception('Implement data module in data.py -> create_data_module()')

        return self.data_module



import torchmetrics

from dsframework.base.trainer.pl_wrapper import *

##
# @file
# @brief generatedClass class, based on pytorch lightning 'LightningModule' base class.
class generatedClass(ZIDSModel):
    """! Model template class inherits from ZIDSModel(LightningModule), which organizes your code into 6 sections:

        - Computations (init).
        - Train Loop (training_step)
        - Validation Loop (validation_step)
        - Test Loop(test_step)
        - Prediction Loop (predict_step)
        - Optimizers and LR Schedulers(configure_optimizers)

        Log parameters:
        - For single value, use:
            self.log("test_loss", test_loss)  # prog_bar=True

        - For multiple values use dictionary, use:
            values = {"loss": loss, "acc": acc, "metric_n": metric_n}  # add more items if needed
            self.log_dict(values)

        View results in tensorboard:
            tensorboard --logdir=lightning_logs/

        Loading datasets:

            example:
            @code{.py}
            train_set = datasets.MNIST(os.getcwd() + '/data', download=True, train=True, transform=transform)
            test_set = datasets.MNIST(os.getcwd() + '/data', download=True, train=False, transform=transform)

            # use 20% of training data for validation
            train_set_size = int(len(train_set) * 0.8)
            valid_set_size = len(train_set) - train_set_size

            # split the train set into two
            seed = torch.Generator().manual_seed(42)
            train_set, valid_set = random_split(train_set, [train_set_size, valid_set_size], generator=seed)

            train_set_loader = DataLoader(train_set, num_workers=2)
            val_set_loader = DataLoader(valid_set, num_workers=2)
            test_set_loader = DataLoader(test_set, num_workers=2)
            @endcode


        Training:
            example:
            @code{.py}
            trainer = pl.Trainer(max_epochs=1, callbacks=[checkpoint_callback])
            trainer.fit(model=autoencoder, train_dataloaders=train_set_loader, val_dataloaders=val_set_loader)
            @endcode

        Testing:
            example:
            @code{.py}
            trainer.test(model=autoencoder, dataloaders=test_set_loader)
            @endcode

        Loading a trained model, use:
            - load_from_checkpoint

                example:
                @code{.py}
                model = MyTrainer.load_from_checkpoint('lightning_logs/epoch=0-step=48000.ckpt')
                @endcode


        Saving a model, use:
            - save_checkpoint

                for example:
                @code{.py}
                trainer.save_checkpoint("my_checkpoint.ckpt")
                @endcode

            - ModelCheckpoint - define automated checkpoint saving, use:
                for example:
                @code{.py}
                checkpoint_callback = ModelCheckpoint(dirpath="lightning_logs/", save_top_k=2, monitor="val_loss")
                @endcode


    """
    def __init__(self, loss_function=F.mse_loss, batch_size=4, num_classes=None, num_classes_conf_matrix=None,
                 learn_rate=1e-3, nnetwork=None, metrics=None):
        """! Model class initializer, receives the network and loss function, it initializes all parameter to be
        tracked. """
        super().__init__(nnetwork=nnetwork, loss_function=None)

        self.nnetwork = nnetwork
        self.loss_function = loss_function
        self.batch_size = batch_size
        self.num_classes = num_classes
        self.num_classes_conf_matrix = num_classes_conf_matrix
        self.learn_rate = learn_rate
        self.metrics = metrics

        for key in self.metrics:
            if not self.metrics[key]:
                continue

            if key == 'accuracy':
                self.train_accuracy = torchmetrics.Accuracy(num_classes=self.num_classes)
                self.val_accuracy = torchmetrics.Accuracy(num_classes=self.num_classes)
                self.test_accuracy = torchmetrics.Accuracy(num_classes=self.num_classes)
            elif key == 'precision':
                self.train_precision = torchmetrics.Precision(num_classes=self.num_classes)
                self.val_precision = torchmetrics.Precision(num_classes=self.num_classes)
                self.test_precision = torchmetrics.Precision(num_classes=self.num_classes)
            elif key == 'recall':
                self.train_recall = torchmetrics.Recall(num_classes=self.num_classes)
                self.val_recall = torchmetrics.Recall(num_classes=self.num_classes)
                self.test_recall = torchmetrics.Recall(num_classes=self.num_classes)
            elif key == 'f1':
                self.train_f1 = torchmetrics.F1Score(num_classes=self.num_classes)
                self.val_f1 = torchmetrics.F1Score(num_classes=self.num_classes)
                self.test_f1 = torchmetrics.F1Score(num_classes=self.num_classes)
            elif key == 'confusion_matrix':
                self.confusion_matrix = torchmetrics.ConfusionMatrix(num_classes=self.num_classes_conf_matrix)

        self.save_hyperparameters()

    def configure_optimizers(self):
        """! Choose what optimizers and learning-rate schedulers to use in your optimization. Normally you’d need one.
                But in the case of GANs or similar you might have multiple.

                    Returns: Any of this 6 options
                        - Single optimizer.
                        - List or Tuple of optimizers.
                        - Two lists - The first list has multiple optimizers, and the second has multiple LR schedulers
                            (or multiple lr_scheduler_config).
                        - Dictionary, with an "optimizer" key, and (optionally) a "lr_scheduler" key whose value
                            is a single LR scheduler or lr_scheduler_config.
                        - Tuple of dictionaries as described above, with an optional "frequency" key.
                        - None - Fit will run without any optimizer.

                Example:
                    @code{.py}
                    optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
                    return optimizer
                    @endcode

                Another example:
                    @code{.py}
                    SGD_optimizer = optim.SGD(model.parameters(), lr=0.001, momentum=0.7)
                    return SGD_optimizer
                    @endcode

        """
        optimizer = Optimizers.Adam(self.parameters(), lr=self.learn_rate)
        return optimizer

    def forward(self, x):
        """! Defines the computation performed at every call.

        Implementation example:
            @code{.py}
            return self.nnetwork(x)
            @endcode
        """
        x = x.view(x.size(0), -1)
        x = self.nnetwork(x)
        return x

    def training_step(self, batch, batch_idx):
        """! Override to enable training loop

        Implementation example:
            @code{.py}
            def training_step(self, batch, batch_idx):
                x, y = batch
                x = x.view(x.size(0), -1)
                z = self.encoder(x)
                x_hat = self.decoder(z)
                loss = function.mse_loss(x_hat, x)
                return loss
            @endcode

        If you need to do something with all the outputs of each training_step(), override the
        training_epoch_end() method.
        """
        x, y = batch
        y_hat = self(x)
        loss = self.loss_function(y_hat, y)
        pred = torch.argmax(y_hat, dim=1)

        self.log_metrics(pred, y, 'train')

        return loss

    def validation_step(self, batch, batch_idx):
        """! Override to enable validation loop

        Implementation example:
            @code{.py}
            def validation_step(self, batch, batch_idx):
                x, y = batch
                x = x.view(x.size(0), -1)
                z = self.encoder(x)
                x_hat = self.decoder(z)
                test_loss = function.mse_loss(x_hat, x)
                self.log("val_loss", test_loss, prog_bar=True)
            @endcode
        """
        x, y = batch
        y_hat = self(x)
        loss = self.loss_function(y_hat, y)
        pred = torch.argmax(y_hat, dim=1)

        self.log_metrics(pred, y, 'val')

        self.log("val_loss", loss, prog_bar=True)

    def test_step(self, batch, batch_idx):
        """! Override to enable test loop.

        Implementation example:
            @code{.py}
            x, y = batch
            x = x.view(x.size(0), -1)
            z = self.encoder(x)
            x_hat = self.decoder(z)
            test_loss = function.mse_loss(x_hat, x)
            self.log("test_loss", test_loss)  # prog_bar=True

            return {'loss': loss, 'preds': pred, 'target': y}  # will be available in test_epoch_end
            @endcode
        """
        x, y = batch
        y_hat = self(x)
        loss = self.loss_function(y_hat, y)
        pred = torch.argmax(y_hat, dim=1)

        self.log_metrics(pred, y, 'test')

        self.log("test_loss", loss)  # prog_bar=True

        return {'loss': loss, 'preds': pred, 'target': y}  # will be available in test_epoch_end

    def test_epoch_end(self, outputs):
        """! Implement this to calculate test results and fill the local dictionary return self.test_results

            Args:
                outputs: Array (ie. dict, list etc.) with accumulated results returned from:
                 def test_step(self, batch, batch_idx)

        Confusion matrix example:
            preds = torch.cat([tmp['preds'] for tmp in outputs])
            targets = torch.cat([tmp['target'] for tmp in outputs])

            conf_matrix_results = self.confusion_matrix(preds, targets)

        Results structure example:
            self.test_results = {
                'precision': precision_result,
                'accuracy': accuracy_result,
                'recall': recall_result,
                'f1': f1_result,
                'confusion_matrix': conf_matrix_results}

        """
        conf_matrix_results = None

        preds = torch.cat([tmp['preds'] for tmp in outputs])
        targets = torch.cat([tmp['target'] for tmp in outputs])

        if self.metrics['confusion_matrix']:
            conf_matrix_results = self.confusion_matrix(preds, targets)

        self.test_results = self.prepare_test_results()

        if self.metrics['confusion_matrix']:
            self.test_results['confusion_matrix'] = conf_matrix_results

    def log_metrics(self, pred, y, step_name):
        """Log the metrics specified in config.py metrics section."""
        for key in self.metrics:

            if not self.metrics[key] or key == 'confusion_matrix':
                continue

            method_name = "%s_%s" % (step_name, key)
            try:
                metric_method = getattr(self, method_name)
            except AttributeError as e:
                print(f'No method {method_name} metric found.')
                return

            val = metric_method(pred, y)
            self.log(method_name, val, prog_bar=True, on_epoch=True)

    def prepare_test_results(self):
        """Prepare test results' dictionary with configured metrics."""
        test_results = {}

        for key in self.metrics:
            if not self.metrics[key] or key == 'confusion_matrix':
                continue

            method_name = "%s_%s" % ('test', key)
            metric_method = getattr(self, method_name)
            test_results[key] = metric_method.compute()

        return test_results

    def get_test_results(self):
        """Returns test_results."""
        return self.test_results



class ZIDSTrainerTest:

    trainer = None
    model = None
    data_module = None

    def __init__(self, data_module=None, model_path=None, trainer_config=None):
        self.data_module = data_module
        self.model_path = model_path

        self.trainer_setup(trainer_config)
        self.load_model(model_path)

    def trainer_setup(self, trainer_config):
        pass

    def load_model(self, model_path):
        pass

    def execute_test(self):
        self.trainer.test(self.model, self.data_module)

        return self.model.get_test_results()



from typing import Any


class ZIDSNetwork:
    """! Base class for neural network definition:

    for example:
        @code
            nnetwork = nn.Sequential(
                nn.Linear(28 * 28, 64),
                nn.ReLU(),
                nn.Linear(64, 3),
            )
        @endcode

    Another example:
        @code
        self.bert = BertModel.from_pretrained("bert-base-cased")
        @endcode
    """
    nnetwork = None

    def __init__(self, nnetwork=None):

        self.nnetwork = nnetwork

        if self.nnetwork is None:
            raise Exception("Network not defined. please define in network_module.py")

    def from_pretrained(self) -> Any:
        pass



import os
from typing import Union, Any
from os.path import exists
import csv
import shutil
from dsframework.base.trainer.pl_wrapper import Dataset, IterableDataset, random_split, transforms, datasets
from comparison import generatedProjectNameComparison
from config import config
from data import generatedProjectNameData
from model import generatedProjectNameModel
from test import generatedProjectNameTest
from train import generatedProjectNameTrain
from tester.general_tester import generatedProjectNameGeneralTester

##
# @file
# @brief TrainerMain class.
class TrainerMain:
    """! Trainer main class, this is where all training components orchestrate.

    Its main job is to run:
    - training - train the model using train dataset and validation dataset.
    - testing - running model against test dataset.
    - eval - running the same test dataset

    Important node:
    It includes a split() method, which needs to be implemented. This is where the dataset gets split to
    train, validation and test datasets.

    """
    general_config = None
    trainer_config = None

    def __init__(self, cfg, dataset: Union[IterableDataset, Dataset, Any] = None):
        """! Trainer main initializer."""
        self.save_path = None
        self.train_class = None
        self.model_class = None
        self.data_class = None
        self.data_class_test = None
        self.test_class = None
        self.general_config = cfg['general_config']
        self.trainer_config = cfg['trainer_config']
        self.metrics_config = cfg['metrics']
        self.dataset = dataset
        self.dataset_train = None
        self.dataset_validation = None
        self.dataset_test = None
        self.eval_res = None
        self.split()

    def split(self):
        """! Implement this method to split between the different datasets:
        For the trainer supply: dataset_train, self.dataset_validation
        For test supply: self.dataset_test

        For example:
            self.dataset = datasets.MNIST(os.getcwd() + '/datasets', download=True, train=True,
                                      transform=transforms.ToTensor())
            self.dataset_train, self.dataset_validation = random_split(self.dataset, [55000, 5000])
            self.dataset_test = datasets.MNIST(os.getcwd() + '/datasets', download=True, train=False,
                                           transform=transforms.ToTensor())
        """

        raise 'Dataset TrainerMain.split() not implemented exception.'

    def save_dataset(self, dataset_path):
        """! Saves a copy of the dataset in /training_output/v1 folder zipped.

        In a later stage the trained model is created in the same folder, so we have a pair of the model and the
        dataset used to create it.
        """
        if not self.general_config['save_dataset']:
            return

        if os.path.isdir(dataset_path):
            if os.listdir(dataset_path):
                shutil.make_archive(os.path.join(self.save_path, 'dataset'), 'zip', dataset_path)

    def execute_trainer(self):
        """ Executes training process. it initializes the main 3 trainer modules:

        generatedProjectNameData - Class for holding datasets and functionality related to dataset handling
        generatedProjectNameModel - Network definition and all functionality related to creating the model.
        generatedProjectNameTrain - Execute trainer and all trainer related functionality.
        """

        self.save_path = self._create_save_path()
        self.save_dataset(os.getcwd() + '/datasets')

        self.data_class = generatedProjectNameData(dataset_train=self.dataset_train,
                                                   dataset_validation=self.dataset_validation,
                                                   save_path=self.save_path)
        self.model_class = generatedProjectNameModel(self.general_config, self.metrics_config)
        self.train_class = generatedProjectNameTrain(
            trainer_config=self.trainer_config,
            general_config=self.general_config,
            data_module=self.data_class.get_data_module(),
            model=self.model_class.get_model(),
            save_path=self.save_path)

        self.train_class.execute()

    def execute_test(self, model_path):
        """ Executes test process. runs the model against a test dataset.

        Initializes:
            generatedProjectNameData - Created with test dataset.
            generatedProjectNameTest - Loads model and executes test.

        Args:
            model_path: Path to saved model
        Returns:
            mmodel_test_results: Metrics - defined in config.
        """

        self.data_class_test = generatedProjectNameData(dataset_test=self.dataset_test)

        self.test_class = generatedProjectNameTest(
            data_module=self.data_class_test.get_data_module(),
            model_path=model_path,
            trainer_config=self.trainer_config)

        model_test_results = self.test_class.execute_test()

        return model_test_results

    def execute_eval(self):
        """ Executes evaluation process on an existing model against a test dataset.

        Notes:
            Test dataset must be in Pandas dataframe format.

        Returns:
            eval_res - Evaluation metrics result.

        Result format example:
            self.eval_res = {'precision': 1.1, 'accuracy': 1.2, 'recall': 1.3, 'f1': 1.4,
                            'confusion_matrix': {'tp': 0, 'tn': 0, 'fp': 0, 'fn': 0}}
        """

        model_tester = generatedProjectNameGeneralTester()

        try:
            dataframe_ds = self.convert_to_dataframe(self.dataset_test)
            output_path = model_tester.test_from_dataset(dataframe_ds, 1)
            self.eval_res = self.get_eval_results(output_path)

        except Exception as e:
            print(f'Could not start evaluation process: {e}')

        return self.eval_res

    def get_eval_results(self, output_path):
        """Reads evaluation results from path and returns it.

        Args:
            output_path - Location of eval results

        Returns:
            evaluation results - In dictionary format.

        """
        if exists(output_path):
            return csv.DictReader(open(output_path))
        else:
            return None

    def convert_to_dataframe(self, dataset_test):
        """Conversion of test dataset to Pandas dataframe, to run evaluation process.

        Not implemented.
        """
        raise "To run evaluation, convert dataset tn pandas dataframe format, implement convert_to_dataframe()."

    def _create_save_path(self):
        """Create the next version folder (v1, v2, v3 ... vn) inside training_output folder.

        Returns:
            save_path - The newly created folder path.
        """

        current_folder = os.path.dirname(os.path.realpath(__file__))

        path = os.path.join(current_folder, 'training_output')

        if not os.path.exists(path):
            os.makedirs(path)

        all_folders = [file for file in os.listdir(path) if os.path.isdir(os.path.join(path, file))]

        if all_folders:
            num_list = [int(folder.replace('v', '')) for folder in all_folders if folder.replace('v', '').isdigit()]
            num_list.sort()
            new_version = num_list[-1] + 1
        else:
            new_version = 1

        save_path = os.path.join(path, f'v{new_version}')
        os.makedirs(save_path)

        return save_path

##
# @file
#
# @brief TrainerMain usage simulation.
if __name__ == '__main__':
    ## TrainerMain for training process
    trainer_main_class = TrainerMain(config, dataset=None)
    trainer_main_class.execute_trainer()

    ## Path to last saved model
    model_path = os.path.join(trainer_main_class.save_path, config['general_config']['model_name'])
    ## TrainerMain for testing process
    trainer_main_class_test = TrainerMain(config, dataset=None)
    test_results = trainer_main_class_test.execute_test(model_path=model_path)
    eval_results = trainer_main_class_test.execute_eval()

    ## The results of the comparison between trained model and existing model evaluation.
    comparison_results = generatedProjectNameComparison(
        test_results=test_results,
        eval_results=eval_results,
        metrics_config=config['metrics']).compare()

    print('comparison_results', comparison_results)



class ZIDSTrainerData:

    dataset_train = None
    dataset_validation = None
    dataset_test = None
    data_module = None
    save_path = ''

    def __init__(self, dataset_train=None, dataset_validation=None, dataset_test=None, save_path=''):
        self.dataset_train = dataset_train
        self.dataset_validation = dataset_validation
        self.dataset_test = dataset_test
        self.save_path = save_path

        self.create_data_module()

    def create_data_module(self):
        pass

    def get_data_module(self):

        self.data_module = None

        return self.data_module



from trainer.data_module import generatedProjectNameDataModule
from trainer.network_module import generatedProjectNameNetworkModule
from trainer.model import generatedProjectNameModel
from dsframework.base.trainer.pl_wrapper import *


def train(config, train_config):
    data_path = '/Users/yuvalbechor/Develop/ds-framework/dsframework/cli/trainer/datasets/wine_quality/wine-quality.csv'

    wine_data_module = generatedProjectNameDataModule(
        data_path=data_path,
        batch_size_train=config['batch_size_train'],
        batch_size_eval=config['batch_size_eval'],
        num_workers=config['num_workers'],
        seed=config['seed'],
        split_percent=0.8,
        test_set_out_of_dataset=True
    )

    model = generatedProjectNameModel(
        loss_function=config['loss_function'],
        num_classes=config['num_classes'],
        learn_rate=config['learning_rate'],
        nnetwork=config['nnetwork']
    )

    # automatically restores model, epoch, step, LR schedulers, apex, etc...
    # trainer.fit(model, ckpt_path="some/path/to/my_checkpoint.ckpt")
    trainer = pl.Trainer(**train_config)
    trainer.fit(model, wine_data_module)

    trainer.test(model, wine_data_module)


if __name__ == '__main__':

    general_config = {
        'learning_rate': 1e-3,
        'loss_function': F.cross_entropy,
        'batch_size_train': 1,
        'batch_size_eval': 1,
        'num_classes': None,
        'num_workers': 1,
        'seed': 42,
        'nnetwork': generatedProjectNameNetworkModule.nnetwork_wine
    }

    trainer_config = {
        'max_epochs': 1,
        'accelerator': None,  # 'cpu', 'gpu', 'tpu', 'ipu', 'auto', None
        'devices': None,  # number of accelerators: -1-all, 1, 3, 4...8, None, [1, 3]-gpu ids, None
        'strategy': None,  # 'dp', 'ddp', 'ddp2', 'ddp_spawn'
        'num_processes': None  # number of cpus
    }

    train(general_config, trainer_config)



from typing import Any

from dsframework.base.trainer.pl_wrapper import *

##
# @file
# @brief generatedClass class, define neural network layers.
class generatedClass(ZIDSNetwork):
    """! Define your networks here, or load from pretrained model.

    for example:
        @code
        nnetwork = nn.Sequential(
            nn.Linear(28 * 28, 64),
            nn.ReLU(),
            nn.Linear(64, 3),
            nn.Linear(3, 64),
            nn.ReLU(),
            nn.Linear(64, 28 * 28)
        )
        @endcode

    Another example:
        @code
        nnetwork_wine = nn.Sequential(
        nn.Linear(11, 1),
        nn.ReLU(),
        nn.Linear(1, 1))
        @endcode

    Another example:
        @code
        self.bert = BertModel.from_pretrained("bert-base-cased")
        @endcode
    """

    nnetwork = None

    if nnetwork is None:
        raise Exception("Network not defined. please define in network_module.py")

    def __init__(self):
        super().__init__(self.nnetwork)

    def from_pretrained(self) -> Any:
        pass



import os
import sys
from typing import List
from google.cloud.dataproc_v1.types.workflow_templates import TemplateParameter
from google.cloud.dataproc_v1.types import OrderedJob
from dsframework.base.batch.workflow_base import ZIDS_Workflow


class Workflow(ZIDS_Workflow):

    def __init__(self):
        """! Loads config file and initializes parameters.
        """
        self.default_jars = ['gs://spark-lib/bigquery/spark-bigquery-latest_2.12.jar']
        main_project_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.."))

        self.pip_packages = ''
        with open(main_project_path + '/requirements_docker.txt') as req_file:
            required_pkg = req_file.read().splitlines()
            required_pkg = [x for x in required_pkg if "#" not in x]
            self.pip_packages = ' '.join(required_pkg)

        super().__init__(self.pip_packages)

        self.whl_pipes_file_name = self.project_name + '_pipes.zip'
        self.whl_no_pipes_file_name = self.project_name + '_no_pipes.zip'

        # Initialize spark properties.
        self.spark_properties = self.config["spark_properties"]

    def create_job(self, stage_name, prerequisite_list=None) -> OrderedJob():
        """! Create a job from a given stage

            Args:
                stage_name: Name of the stage to create
                prerequisite_list: list of stage id's that are mandatory to run before current stage

            Returns:
                An Ordered Job, to be inserted into the workflow template
        """
        stage_dir = f'batch_files/stages/{stage_name}'
        python_files = [self.whl_pipes_file_name,
                        self.whl_no_pipes_file_name]
        archive_files = [self.whl_pipes_file_name,
                         self.whl_no_pipes_file_name]
        stage_job = self.create_ordered_job(stage_dir, self.spark_properties, python_files, archive_files)

        if prerequisite_list is not None:
            stage_job.prerequisite_step_ids = prerequisite_list

        return stage_job

    def create_stages(self) -> List[OrderedJob]:
        """! Create the stages for the workflow stages
            Returns:
                A List of Ordered Jobs, to be inserted into the workflow template
        """
        # stage get_data
        stage_get_data_job = self.create_job("get_data")

        # stage run_pipeline
        prerequisite_list = [stage_get_data_job.step_id]
        stage_run_pipeline_job = self.create_job("run_pipeline", prerequisite_list)

        # stage set_data
        prerequisite_list = [stage_run_pipeline_job.step_id]
        stage_set_data_job = self.create_job("set_data", prerequisite_list)

        return [stage_get_data_job, stage_run_pipeline_job, stage_set_data_job]

    def finalize_jobs(self, workflow_template):
        """! Finalize the jobs, introduce additional parameters, etc.
            Args:
                workflow_template: Workflow templates. Includes jobs that we wish to modify
        """

        # Make sure to change jobs according to specific configuration
        if self.config['specific_jobs_to_run'] and len(self.config['specific_jobs_to_run']):
            specific_jobs_to_run = []
            for job in workflow_template.jobs:
                if job.step_id in self.config['specific_jobs_to_run']:
                    specific_jobs_to_run.append(job)
                    if self.config['force_remove_all_prerequisite']:
                        job.prerequisite_step_ids = []
                    if len(job.prerequisite_step_ids) and not all(
                            elem in specific_jobs_to_run for elem in job.prerequisite_step_ids):
                        raise Exception("job depends on another job not in specific_jobs_to_run list")
            workflow_template.jobs = specific_jobs_to_run

        # Set start and end dates from Airflow composer
        workflow_template.parameters = []

        start_date_param = TemplateParameter()
        start_date_param.description = "start date from airflow composer"
        start_date_param.name = "START_DATE"

        end_date_param = TemplateParameter()
        end_date_param.description = "end date from airflow composer"
        end_date_param.name = "END_DATE"

        extra_param = TemplateParameter()
        extra_param.description = "allow adding extra parameters"
        extra_param.name = "PARAMS"

        for job in workflow_template.jobs:
            start_date_param.fields.append("jobs['" + job.step_id + "'].pysparkJob.args[1]")
            end_date_param.fields.append("jobs['" + job.step_id + "'].pysparkJob.args[2]")
            extra_param.fields.append("jobs['" + job.step_id + "'].pysparkJob.args[3]")

        workflow_template.parameters.append(start_date_param)
        workflow_template.parameters.append(end_date_param)
        workflow_template.parameters.append(extra_param)

    def export_template_structure(self):
        """! Create the workflow template, and return it for external usage
            Returns:
                workflow_template: A ready to use template, WorkflowTemplate() class
        """
        # Define template structure
        workflow_template = self.create_template_structure()
        workflow_template.jobs = self.create_stages()
        self.finalize_jobs(workflow_template)
        return workflow_template

    def create_workflow_template(self, create_yaml: bool = False):
        """! Create the workflow template, along with all relevant stages
            Args:
                create_yaml: Boolean, deciding whether we should create a yaml file out of the template
        """

        # Check if a cluster exists. Not mandatory, as the workflow can create a cluster on the fly
        self.check_cluster()

        # Define template structure
        workflow_template = self.export_template_structure()

        print(f'{workflow_template}')

        # Make sure that the template does not exist
        self.check_template()

        # If needed, create a .yaml file, don't use Cloud API
        if create_yaml:
            self.create_yaml_file(workflow_template)

        else:
            # Create template by using Cloud API
            self.create_template(workflow_template)


if __name__ == "__main__":
    """! Triggers the creation of the workflow template, starting from
    basic configuration, defining and creating jobs from the existing stages, 
    and up to creating and instantiating the workflow, if needed.
        Args:
        System argument 1 - Action to be performed on a template (create, delete, instantiate)
    """

    template_action = "none"
    if sys.argv and len(sys.argv) > 1:
        template_action = sys.argv[1]
    workflow = Workflow()

    if template_action == 'create':
        workflow.create_workflow_template()

    if template_action == 'create_yaml':
        workflow.create_workflow_template(True)

    # At this point, a template is either previously or recently created
    template = workflow.get_workflow_template()

    if template_action == 'delete' and template:
        workflow.delete_workflow_template(template)

    if template_action == 'instantiate' and template:
        workflow.instantiate_template(template)



from data_module import generatedProjectNameDataModule
from network_module import generatedProjectNameNetworkModule
from plmodel import generatedProjectNameModel
from dsframework.base.trainer.pl_wrapper import *

##
# Steps to work with the trainer.py:
#
# 1. Datasets:
#   a. Load dataset
#   b. Parse it - override parse_dataset() method in data_module.py.
#
# 2. Model:
#   a. Define your network in NetworkModule class in network.py.
#   b. Check model methods in model.py.
#
# 3. Callbacks:
#   a. Checkpoint - Adjust ModelCheckpoint callback parameters in this file - trainer.py.
#   b. Early stopping - Adjust EarlyStopping callback parameters in this file - trainer.py.
#
# 4. Configuration:
#   a. Adjust general config in this file - trainer.py
#   b. Adjust trainer config in this file - trainer.py


def train(config, train_config):
    """Train method to be used only if using pytorch lightning directly and not via the trainer main classes."""
    # either load the dataset,
    full_dataset = None
    test_dataset = None
    # or provide a path
    data_path = './path to dataset'
    file_type = 'csv'

    data_module = generatedProjectNameDataModule(
        data_path=data_path,
        file_type=file_type,
        batch_size_train=config['batch_size_train'],
        batch_size_eval=config['batch_size_eval'],
        dataset=full_dataset,
        test_set=test_dataset,
        num_workers=config['num_workers'],
        seed=config['seed'],
        split_percent=0.9,  # Split between train and validation datasets
        test_set_out_of_dataset=False  # Split a test data out of the dataset.
    )

    model = generatedProjectNameModel(
        loss_function=config['loss_function'],
        num_classes=config['num_classes'],
        learn_rate=config['learning_rate'],
        nnetwork=config['nnetwork']
    )

    # To automatically restores model, epoch, step, LR schedulers, apex, etc... use the following:
    # trainer.fit(model, ckpt_path="some/path/to/my_checkpoint.ckpt")
    trainer = pl.Trainer(**train_config)
    trainer.fit(model, data_module)

    trainer.test(model, data_module)


# Define checkpoint callback
# saves top-K checkpoints based on "val_loss" metric
checkpoint_callback = ModelCheckpoint(
    save_top_k=3,
    save_last=True,
    monitor="val_loss",
    mode="min",
    filename="my-chkpnt-{epoch:02d}-{val_loss:.2f}",
)

early_stopping_callback = EarlyStopping(
    monitor='val_loss',
    patience=10,  # number of val checks with no improvement
    mode='max'
)

if __name__ == '__main__':

    general_config = {
        'learning_rate': 1e-3,
        'loss_function': F.cross_entropy,
        'batch_size_train': 32,
        'batch_size_eval': 32,
        'num_classes': None,
        'num_workers': 1,  # Number of workers assigned to DataLoader (number of cpus)
        'seed': 42,
        'nnetwork': generatedProjectNameNetworkModule.nnetwork
    }

    trainer_config = {
        'max_epochs': 1,
        'accelerator': None,  # 'cpu', 'gpu', 'tpu', 'ipu', 'auto', None
        'devices': None,  # number of accelerators: -1-all, 1, 3, 4...8, None, [1, 3]-gpu ids, None
        'strategy': None,  # 'dp', 'ddp', 'ddp2', 'ddp_spawn'
        'num_processes': None,  # number of cpus
        'enable_checkpointing': True,
        'default_root_dir': './',  # root dir to save checkpoints
        'callbacks': [early_stopping_callback, checkpoint_callback],
        'enable_progress_bar': False  # Set to 'True' to see logs and progress bar in console
    }

    train(general_config, trainer_config)
